#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2016 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

import datetime

from django.conf import settings
from django.contrib.gis.db import models
from django.core.urlresolvers import reverse
from django.db import connection, transaction
from django.db.models import Max, Q
from django.db.models.signals import m2m_changed, post_save, post_delete
from django.utils.translation import ugettext_lazy as _, ugettext

from ishtar_common.utils import cached_label_changed, post_save_point

from ishtar_common.models import GeneralType, ImageModel, BaseHistorizedItem, \
    ShortMenuItem, LightHistorizedItem, HistoricalRecords, OwnPerms, Source, \
    Person, Basket, get_external_id, post_save_cache, ValueGetter

from archaeological_operations.models import AdministrativeAct
from archaeological_context_records.models import ContextRecord, Dating

from ishtar_common.models import PRIVATE_FIELDS, SpatialReferenceSystem


class MaterialType(GeneralType):
    code = models.CharField(_(u"Code"), max_length=10, blank=True, null=True)
    recommendation = models.TextField(_(u"Recommendation"), blank=True,
                                      null=True)
    parent = models.ForeignKey("MaterialType", blank=True, null=True,
                               verbose_name=_(u"Parent material"))

    class Meta:
        verbose_name = _(u"Material type")
        verbose_name_plural = _(u"Material types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=MaterialType)
post_delete.connect(post_save_cache, sender=MaterialType)


class ConservatoryState(GeneralType):
    parent = models.ForeignKey("ConservatoryState", blank=True, null=True,
                               verbose_name=_(u"Parent conservatory state"))

    class Meta:
        verbose_name = _(u"Conservatory state")
        verbose_name_plural = _(u"Conservatory states")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=ConservatoryState)
post_delete.connect(post_save_cache, sender=ConservatoryState)


class PreservationType(GeneralType):
    class Meta:
        verbose_name = _(u"Preservation type")
        verbose_name_plural = _(u"Preservation types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=PreservationType)
post_delete.connect(post_save_cache, sender=PreservationType)


class IntegrityType(GeneralType):
    class Meta:
        verbose_name = _(u"Integrity / interest type")
        verbose_name_plural = _(u"Integrity / interest types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=IntegrityType)
post_delete.connect(post_save_cache, sender=IntegrityType)


class RemarkabilityType(GeneralType):
    class Meta:
        verbose_name = _(u"Remarkability type")
        verbose_name_plural = _(u"Remarkability types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=RemarkabilityType)
post_delete.connect(post_save_cache, sender=RemarkabilityType)


class BatchType(GeneralType):
    order = models.IntegerField(_(u"Order"), default=10)
    class Meta:
        verbose_name = _(u"Batch type")
        verbose_name_plural = _(u"Batch types")
        ordering = ('order',)
post_save.connect(post_save_cache, sender=BatchType)
post_delete.connect(post_save_cache, sender=BatchType)


class ObjectType(GeneralType):
    parent = models.ForeignKey("ObjectType", blank=True, null=True,
                               verbose_name=_(u"Parent"))

    class Meta:
        verbose_name = _(u"Object type")
        verbose_name_plural = _(u"Object types")
        ordering = ('parent__label', 'label',)

    def full_label(self):
        lbls = [self.label]
        item = self
        while item.parent:
            item = item.parent
            lbls.append(item.label)
        return u" > ".join(reversed(lbls))

    def __unicode__(self):
        return self.label
post_save.connect(post_save_cache, sender=ObjectType)
post_delete.connect(post_save_cache, sender=ObjectType)


class BFBulkView(object):
    CREATE_SQL = """
        CREATE VIEW basefind_cached_bulk_update
        AS (
            SELECT bf.id AS id, ope.code_patriarche AS main_ope_code,
                ope.year AS year,
                ope.operation_code AS ope_code,
                cr.label AS cr_label,
                bf.index AS index
            FROM archaeological_finds_basefind bf
            INNER JOIN archaeological_context_records_contextrecord cr
                ON cr.id = bf.context_record_id
            INNER JOIN archaeological_operations_operation ope
                ON ope.id = cr.operation_id
        );"""
    DELETE_SQL = """
        DROP VIEW basefind_cached_bulk_update;
    """


class BaseFind(BaseHistorizedItem, OwnPerms):
    label = models.TextField(_(u"Free ID"))
    external_id = models.TextField(_(u"External ID"), blank=True, null=True)
    auto_external_id = models.BooleanField(
        _(u"External ID is set automatically"), default=False)
    description = models.TextField(_(u"Description"), blank=True, null=True)
    comment = models.TextField(_(u"Comment"), blank=True, null=True)
    special_interest = models.CharField(_(u"Special interest"), blank=True,
                                        null=True, max_length=120)
    context_record = models.ForeignKey(
        ContextRecord, related_name='base_finds',
        verbose_name=_(u"Context Record"))
    discovery_date = models.DateField(_(u"Discovery date"),
                                      blank=True, null=True)
    batch = models.ForeignKey(BatchType, verbose_name=_(u"Batch/object"),
                              blank=True, null=True)
    index = models.IntegerField(u"Index", default=0)
    material_index = models.IntegerField(_(u"Material index"), default=0)
    topographic_localisation = models.CharField(
        _(u"Point of topographic reference"), blank=True, null=True,
        max_length=120)
    x = models.FloatField(_(u'X'), blank=True, null=True)
    y = models.FloatField(_(u'Y'), blank=True, null=True)
    z = models.FloatField(_(u'Z'), blank=True, null=True)
    estimated_error_x = models.FloatField(_(u'Estimated error for X'),
                                          blank=True, null=True)
    estimated_error_y = models.FloatField(_(u'Estimated error for Y'),
                                          blank=True, null=True)
    estimated_error_z = models.FloatField(_(u'Estimated error for Z'),
                                          blank=True, null=True)
    spatial_reference_system = models.ForeignKey(
        SpatialReferenceSystem, verbose_name=_(u"Spatial Reference System"),
        blank=True, null=True)
    point_2d = models.PointField(_(u"Point (2D)"), blank=True, null=True)
    point = models.PointField(_(u"Point"), blank=True, null=True, dim=3)
    line = models.LineStringField(_(u"Line"), blank=True, null=True)
    polygon = models.PolygonField(_(u"Polygon"), blank=True, null=True)
    cache_short_id = models.TextField(
        _(u"Short ID"), blank=True, null=True,
        help_text=_(u"Cached value - do not edit"))
    cache_complete_id = models.TextField(
        _(u"Complete ID"), blank=True, null=True,
        help_text=_(u"Cached value - do not edit"))
    history = HistoricalRecords()
    RELATED_POST_PROCESS = ['find']
    CACHED_LABELS = ['cache_short_id', 'cache_complete_id']

    class Meta:
        verbose_name = _(u"Base find")
        verbose_name_plural = _(u"Base finds")
        permissions = (
            ("view_basefind", ugettext(u"Can view all Base finds")),
            ("view_own_basefind", ugettext(u"Can view own Base find")),
            ("add_own_basefind", ugettext(u"Can add own Base find")),
            ("change_own_basefind", ugettext(u"Can change own Base find")),
            ("delete_own_basefind", ugettext(u"Can delete own Base find")),
        )

    def __unicode__(self):
        return self.label

    def get_last_find(self):
        # TODO: manage virtuals - property(last_find) ?
        finds = self.find.filter().order_by("-order").all()
        return finds and finds[0]

    @classmethod
    def get_max_index(cls, operation):
        q = BaseFind.objects \
            .filter(context_record__operation=operation)
        if q.count():
            return q.aggregate(Max('index'))['index__max']
        return 0

    def _ope_code(self):
        if not self.context_record.operation:
            return ''
        ope = self.context_record.operation
        c_id = []
        if ope.code_patriarche:
            c_id.append(settings.ISHTAR_OPE_PREFIX +
                        unicode(ope.code_patriarche))
        elif ope.year and ope.operation_code:
            c_id.append(
                settings.ISHTAR_DEF_OPE_PREFIX +
                unicode(ope.year or '') + u"-" +
                unicode(ope.operation_code or ''))
        else:
            c_id.append('')
        return settings.JOINT.join(c_id)

    def complete_id(self):
        # OPE|MAT.CODE|UE|FIND_index
        c_id = [self._ope_code()]

        materials = set()
        for find in self.find.filter(downstream_treatment__isnull=True):
            for mat in find.material_types.all():
                if mat.code:
                    materials.add(mat.code)
        c_id.append(u'-'.join(sorted(list(materials))))

        c_id.append(self.context_record.label)

        c_id.append((u'{:0' + str(settings.ISHTAR_FINDS_INDEX_ZERO_LEN) + 'd}'
                     ).format(self.index))
        return settings.JOINT.join(c_id)

    def _generate_cache_complete_id(self):
        return self.complete_id()

    def short_id(self):
        # OPE|FIND_index
        c_id = [self._ope_code()]
        c_id.append((u'{:0' + str(settings.ISHTAR_FINDS_INDEX_ZERO_LEN) + 'd}'
                     ).format(self.index))
        return settings.JOINT.join(c_id)

    def _generate_cache_short_id(self):
        return self.short_id()

    def full_label(self):
        return self._real_label() or self._temp_label() or u""

    def material_type_label(self):
        find = self.get_last_find()
        finds = [find and find.material_type.code or '']
        ope = self.context_record.operation
        finds += [unicode(ope.code_patriarche) or
                  (unicode(ope.year) + "-" + unicode(ope.operation_code))]
        finds += [self.context_record.label, unicode(self.material_index)]
        return settings.JOINT.join(finds)

    def _real_label(self):
        if not self.context_record.parcel \
                or not self.context_record.operation \
                or not self.context_record.operation.code_patriarche:
            return
        find = self.get_last_find()
        lbl = find.label or self.label
        return settings.JOINT.join(
            [unicode(it) for it in (
                self.context_record.operation.code_patriarche,
                self.context_record.label, lbl) if it])

    def _temp_label(self):
        if not self.context_record.parcel:
            return
        find = self.get_last_find()
        lbl = find.label or self.label
        return settings.JOINT.join(
            [unicode(it) for it in (
                self.context_record.parcel.year, self.index,
                self.context_record.label, lbl) if it])

    @property
    def name(self):
        return self.label

    @classmethod
    def get_extra_fields(cls):
        fields = {}
        for field in Find._meta.many_to_many:
            if field.name == 'base_finds':
                fields['find'] = field.related.model
        return fields

    def save(self, *args, **kwargs):
        returned = super(BaseFind, self).save(*args, **kwargs)

        updated = False
        if not self.external_id or self.auto_external_id:
            external_id = get_external_id('base_find_external_id', self)
            if external_id != self.external_id:
                updated = True
                self.auto_external_id = True
                self.external_id = external_id
        if updated:
            self._cached_label_checked = False
            self.save()
        return returned

    @classmethod
    def cached_label_bulk_update(cls, operation_id=None, parcel_id=None,
                                 context_record_id=None):
        if operation_id:
            filters = """
            INNER JOIN archaeological_context_records_contextrecord acr
            ON acr.operation_id = %s AND acr.id = mybf.context_record_id
            """
            args = [int(operation_id)]
            kwargs = {'operation_id': operation_id}
        elif parcel_id:
            filters = """
            INNER JOIN archaeological_context_records_contextrecord acr
            ON acr.parcel_id = %s AND acr.id = mybf.context_record_id
            """
            args = [int(parcel_id)]
            kwargs = {'parcel_id': parcel_id}
        elif context_record_id:
            filters = """
            WHERE mybf.context_record_id = %s
            """
            args = [int(context_record_id)]
            kwargs = {'context_record_id': context_record_id}
        else:
            return

        sql = """
        UPDATE "archaeological_finds_basefind" AS bf
            SET cache_short_id =
                CASE
                WHEN basefind_cached_bulk_update.main_ope_code is NULL
                THEN
                    CASE
                    WHEN basefind_cached_bulk_update.year
                            IS NOT NULL
                        AND basefind_cached_bulk_update.ope_code
                            IS NOT NULL
                    THEN
                       '{ope_prefix}' ||
                       basefind_cached_bulk_update.year ||
                       '-' ||
                       basefind_cached_bulk_update.ope_code
                    ELSE ''
                    END
                ELSE
                    '{main_ope_prefix}' ||
                    basefind_cached_bulk_update.main_ope_code
                END
                || '{join}' ||
                to_char(basefind_cached_bulk_update.index, 'fm{zeros}'),

            cache_complete_id =
                CASE
                WHEN basefind_cached_bulk_update.main_ope_code IS NULL
                THEN
                    CASE
                    WHEN basefind_cached_bulk_update.year
                            IS NOT NULL
                        AND basefind_cached_bulk_update.ope_code
                            IS NOT NULL
                    THEN
                       '{ope_prefix}' ||
                       basefind_cached_bulk_update.year ||
                       '-' ||
                       basefind_cached_bulk_update.ope_code
                    ELSE ''
                    END
                ELSE
                    '{main_ope_prefix}' ||
                    basefind_cached_bulk_update.main_ope_code
                END
                || '{join}' ||

                COALESCE(
                (SELECT string_agg(code, '-') FROM
                 (SELECT DISTINCT mt.code AS code FROM
                    archaeological_finds_find_material_types fmt
                    INNER JOIN archaeological_finds_find f
                     ON f.id=fmt.find_id AND f.downstream_treatment_id IS NULL
                    INNER JOIN find_first_base_find fbf
                     ON fbf.find_id = f.id AND
                        basefind_cached_bulk_update.id = fbf.basefind_id
                    INNER JOIN archaeological_finds_materialtype mt
                     ON mt.id = fmt.materialtype_id ORDER BY mt.code)
                  as ag),
                 '-')

                || '{join}' ||
                basefind_cached_bulk_update.cr_label
                || '{join}' ||
                to_char(basefind_cached_bulk_update.index, 'fm{zeros}')

            FROM basefind_cached_bulk_update
            WHERE bf.id = basefind_cached_bulk_update.id
                    AND bf.id IN (
                SELECT mybf.id FROM archaeological_finds_basefind mybf
                {filters}
            );
        """.format(main_ope_prefix=settings.ISHTAR_OPE_PREFIX,
                   ope_prefix=settings.ISHTAR_DEF_OPE_PREFIX,
                   join=settings.JOINT, filters=filters,
                   zeros=settings.ISHTAR_FINDS_INDEX_ZERO_LEN * "0")
        # with connection.cursor() as c:  # django 1.8
        c = connection.cursor()
        c.execute(sql, args)
        transaction.commit_unless_managed()
        cls._meta.get_field_by_name(
            'find')[0].model.cached_label_bulk_update(**kwargs)

post_save.connect(post_save_point, sender=BaseFind)


WEIGHT_UNIT = (('g', _(u"g")),
               ('kg', _(u"kg")),)

CHECK_CHOICES = (('NC', _(u"Not checked")),
                 ('CI', _(u"Checked but incorrect")),
                 ('CC', _(u"Checked and correct")),
                 )


class FindBasket(Basket):
    items = models.ManyToManyField('Find', blank=True, null=True,
                                   related_name='basket')


class FirstBaseFindView(object):
    CREATE_SQL = """
    CREATE VIEW find_first_base_find
    AS (
        SELECT find_id, min(basefind_id) as basefind_id
            FROM archaeological_finds_find_base_finds GROUP BY find_id
    );"""
    DELETE_SQL = """
        DROP VIEW find_first_base_find;
    """


class FBulkView(object):
    CREATE_SQL = """
        CREATE VIEW find_cached_bulk_update
        AS (
            SELECT f.id AS id, ope.code_patriarche AS main_ope_code,
                ope.year AS year,
                ope.operation_code AS ope_code,
                f.label AS label,
                f.index AS index
            FROM archaeological_finds_find f
            INNER JOIN find_first_base_find fbf
                ON fbf.find_id = f.id
            INNER JOIN archaeological_finds_basefind bf
                ON fbf.basefind_id = bf.id
            INNER JOIN archaeological_context_records_contextrecord cr
                ON cr.id = bf.context_record_id
            INNER JOIN archaeological_operations_operation ope
                ON ope.id = cr.operation_id
        );"""
    DELETE_SQL = """
        DROP VIEW find_cached_bulk_update;
    """


class Find(ValueGetter, BaseHistorizedItem, ImageModel, OwnPerms,
           ShortMenuItem):
    CHECK_DICT = dict(CHECK_CHOICES)
    SHOW_URL = 'show-find'
    SLUG = 'find'
    TABLE_COLS = ['label', 'base_finds__context_record__parcel__town',
                  'base_finds__context_record__operation__common_name',
                  'base_finds__context_record__parcel',
                  'base_finds__context_record__label',
                  'material_types__label', 'object_types',
                  'datings__period__label',
                  'container__cached_label',
                  'base_finds__batch', ]
    if settings.COUNTRY == 'fr':
        TABLE_COLS.insert(
            2, 'base_finds__context_record__operation__code_patriarche')
    TABLE_COLS_FOR_OPE = [
        'base_finds__cache_short_id',
        'base_finds__cache_complete_id',
        'previous_id', 'label', 'material_types__label',
        'datings__period__label', 'find_number', 'object_types',
        'container__cached_label',
        'description',
        'base_finds__context_record__parcel__town',
        'base_finds__context_record__parcel', ]
    COL_LABELS = {
        'base_finds__context_record__label': _(u"Context record"),
        'base_finds__cache_short_id': _(u"Base find - Short ID"),
        'base_finds__cache_complete_id': _(u"Base find - Complete ID"),
        'base_finds__context_record__operation__code_patriarche': _(
            u"Operation (code)"
        ),
        'base_finds__context_record__parcel__town': _(u"Town"),
        'base_finds__context_record__operation__common_name': _(
            u"Operation (name)"
        ),
        'base_finds__context_record__parcel': _(u"Parcel"),
        'base_finds__batch':_(u"Batch"),
        'base_finds__comment': _(u"Base find - Comment"),
        'base_finds__description': _(u"Base find - Description"),
        'base_finds__topographic_localisation': _(u"Base find - "
                                                  u"Topographic localisation"),
        'base_finds__special_interest': _(u"Base find - Special interest"),
        'base_finds__discovery_date': _(u"Base find - Discovery date"),
        'container__cached_label': _(u"Container"),
        'datings__period__label': _(u"Periods"),
        'material_types__label': _(u"Material types"),
    }

    EXTRA_FULL_FIELDS = [
        'base_finds__cache_short_id', 'base_finds__cache_complete_id',
        'base_finds__comment', 'base_finds__description',
        'base_finds__topographic_localisation',
        'base_finds__special_interest',
        'base_finds__discovery_date']
    ATTRS_EQUIV = {'get_first_base_find': 'base_finds'}

    # search parameters
    REVERSED_BOOL_FIELDS = ['image__isnull']
    RELATION_TYPES_PREFIX = {
        'ope_relation_types':
            'base_finds__context_record__operation__'}
    RELATIVE_SESSION_NAMES = [
        ('contextrecord', 'base_finds__context_record__pk'),
        ('operation', 'base_finds__context_record__operation__pk'),
        ('file', 'base_finds__context_record__operation__associated_file__pk')
    ]
    BASE_REQUEST = {'downstream_treatment__isnull': True}
    EXTRA_REQUEST_KEYS = {
        'base_finds__cache_short_id':
            'base_finds__cache_short_id__icontains',
        'base_finds__cache_complete_id':
            'base_finds__cache_complete_id__icontains',
        'label':
            'label__icontains',
        'base_finds__context_record':
            'base_finds__context_record__pk',
        'base_finds__context_record__parcel__town':
            'base_finds__context_record__parcel__town',
        'base_finds__context_record__operation__year':
            'base_finds__context_record__operation__year__contains',
        'base_finds__context_record__operation':
            'base_finds__context_record__operation__pk',
        'archaeological_sites':
            'base_finds__context_record__operation__archaeological_sites__pk',
        'base_finds__context_record__operation__code_patriarche':
            'base_finds__context_record__operation__code_patriarche',
        'datings__period': 'datings__period__pk',
        'material_types': 'material_types__pk',
        'base_finds__find__description':
            'base_finds__find__description__icontains',
        'base_finds__batch': 'base_finds__batch',
        'basket': 'basket',
        'cached_label': 'cached_label__icontains',
        'image__isnull': 'image__isnull',
        'container__location': 'container__location__pk',
        'container__responsible': 'container__responsible__pk',
        'container__index': 'container__index',
        'container__reference': 'container__reference',
    }
    EXTRA_REQUEST_KEYS.update(
        dict(
            [(key, key) for key in TABLE_COLS
             if key not in EXTRA_REQUEST_KEYS]
        )
    )
    EXTRA_REQUEST_KEYS.update(
        dict(
            [(key, key) for key in TABLE_COLS_FOR_OPE
             if key not in EXTRA_REQUEST_KEYS]
        )
    )

    # fields
    base_finds = models.ManyToManyField(BaseFind, verbose_name=_(u"Base find"),
                                        related_name='find')
    external_id = models.TextField(_(u"External ID"), blank=True, null=True)
    auto_external_id = models.BooleanField(
        _(u"External ID is set automatically"), default=False)
    order = models.IntegerField(_(u"Order"), default=1)
    label = models.TextField(_(u"Free ID"))
    description = models.TextField(_(u"Description"), blank=True, null=True)
    material_types = models.ManyToManyField(
        MaterialType, verbose_name=_(u"Material types"), related_name='finds')
    conservatory_state = models.ForeignKey(
        ConservatoryState, verbose_name=_(u"Conservatory state"), blank=True,
        null=True, on_delete=models.SET_NULL)
    conservatory_comment = models.TextField(_(u"Conservatory comment"),
                                            blank=True, null=True)
    preservation_to_considers = models.ManyToManyField(
        PreservationType, verbose_name=_(u"Type of preservation to consider"),
        related_name='finds')
    volume = models.FloatField(_(u"Volume (l)"), blank=True, null=True)
    weight = models.FloatField(_(u"Weight (g)"), blank=True, null=True)
    weight_unit = models.CharField(_(u"Weight unit"), max_length=4,
                                   blank=True, null=True, choices=WEIGHT_UNIT)
    find_number = models.IntegerField(_(u"Find number"), blank=True, null=True)
    upstream_treatment = models.ForeignKey(
        "Treatment", blank=True, null=True,
        related_name='downstream', on_delete=models.SET_NULL,
        verbose_name=_(u"Upstream treatment"))
    downstream_treatment = models.ForeignKey(
        "Treatment", blank=True, null=True, related_name='upstream',
        verbose_name=_(u"Downstream treatment"), on_delete=models.SET_NULL)
    datings = models.ManyToManyField(Dating, verbose_name=_(u"Dating"),
                                     related_name='find')
    container = models.ForeignKey(
        "archaeological_warehouse.Container", verbose_name=_(u"Container"),
        blank=True, null=True,
        related_name='finds', on_delete=models.SET_NULL)
    is_complete = models.NullBooleanField(_(u"Is complete?"), blank=True,
                                          null=True)
    object_types = models.ManyToManyField(
        ObjectType, verbose_name=_(u"Object types"), related_name='find')
    integrities = models.ManyToManyField(
        IntegrityType, verbose_name=_(u"Integrity / interest"),
        related_name='find')
    remarkabilities = models.ManyToManyField(
        RemarkabilityType, verbose_name=_(u"Remarkability"),
        related_name='find')
    min_number_of_individuals = models.IntegerField(
        _(u"Minimum number of individuals (MNI)"), blank=True, null=True)
    length = models.FloatField(_(u"Length (cm)"), blank=True, null=True)
    width = models.FloatField(_(u"Width (cm)"), blank=True, null=True)
    height = models.FloatField(_(u"Height (cm)"), blank=True, null=True)
    diameter = models.FloatField(_(u"Diameter (cm)"), blank=True, null=True)
    thickness = models.FloatField(_(u"Thickness (cm)"), blank=True, null=True)
    dimensions_comment = models.TextField(_(u"Dimensions comment"),
                                          blank=True, null=True)
    mark = models.TextField(_(u"Mark"), blank=True, null=True)
    comment = models.TextField(_(u"Comment"), blank=True, null=True)
    dating_comment = models.TextField(_(u"Comment on dating"), blank=True,
                                      null=True)
    previous_id = models.TextField(_(u"Previous ID"), blank=True, null=True)
    index = models.IntegerField(u"Index", default=0)
    checked = models.CharField(_(u"Check"), max_length=2, default='NC',
                               choices=CHECK_CHOICES)
    check_date = models.DateField(_(u"Check date"),
                                  default=datetime.date.today)
    estimated_value = models.FloatField(_(u"Estimated value"), blank=True,
                                        null=True)
    collection = models.ForeignKey(
        "archaeological_warehouse.Collection", verbose_name=_(u"Collection"),
        blank=True, null=True, related_name='finds', on_delete=models.SET_NULL)
    cached_label = models.TextField(_(u"Cached name"), null=True, blank=True)
    history = HistoricalRecords()
    BASKET_MODEL = FindBasket
    IMAGE_PREFIX = 'finds/'

    class Meta:
        verbose_name = _(u"Find")
        verbose_name_plural = _(u"Finds")
        permissions = (
            ("view_find", ugettext(u"Can view all Finds")),
            ("view_own_find", ugettext(u"Can view own Find")),
            ("add_own_find", ugettext(u"Can add own Find")),
            ("change_own_find", ugettext(u"Can change own Find")),
            ("delete_own_find", ugettext(u"Can delete own Find")),
        )
        ordering = ('cached_label',)

    @property
    def short_class_name(self):
        return _(u"FIND")

    def __unicode__(self):
        lbl = settings.JOINT.join(
            [getattr(self, attr) for attr in ('administrative_index', 'label')
             if getattr(self, attr)])
        return lbl

    @property
    def short_label(self):
        return self.reference

    @property
    def dating(self):
        return u" ; ".join([unicode(dating) for dating in self.datings.all()])

    @property
    def show_url(self):
        return reverse('show-find', args=[self.pk, ''])

    @property
    def name(self):
        return u" - ".join([base_find.name
                            for base_find in self.base_finds.all()])

    @property
    def full_label(self):
        lbl = u" - ".join([getattr(self, attr)
                           for attr in ('label', 'administrative_index')
                           if getattr(self, attr)])
        base = u" - ".join([base_find.complete_id()
                            for base_find in self.base_finds.all()])
        if base:
            lbl += u' ({})'.format(base)
        return lbl

    def get_first_base_find(self):
        q = self.base_finds
        if not q.count():
            return
        return q.order_by('-pk').all()[0]

    @property
    def reference(self):
        bf = self.get_first_base_find()
        if not bf:
            return "00"
        return bf.short_id()

    @property
    def administrative_index(self):
        bf = self.get_first_base_find()
        if not bf or not bf.context_record or not bf.context_record.operation:
            return ""
        return "{}-{}".format(
            bf.context_record.operation.get_reference(),
            self.index)

    def _get_treatments(self, model, rel='upstream', limit=None):
        treatments, findtreats = [], []
        q = model.objects.filter(
            find_id=self.pk).order_by(
            'treatment_nb', 'treatment__start_date',
            'treatment__end_date')
        for findtreat in q.distinct().all():
            if findtreat.pk in findtreats:
                continue
            findtreats.append(findtreat.pk)
            q = getattr(findtreat.treatment, rel).distinct().order_by(
                'label')
            if limit:
                q = q[:limit]
            treatments.append((q.all(), findtreat.treatment))
        return treatments

    @property
    def weight_string(self):
        if not self.weight:
            return ""
        return "{} {}".format(self.weight, self.weight_unit or "")

    def upstream_treatments(self, limit=None):
        from archaeological_finds.models_treatments import \
            FindUpstreamTreatments
        return self._get_treatments(FindUpstreamTreatments, 'upstream',
                                    limit=limit)

    def limited_upstream_treatments(self):
        return self.upstream_treatments(15)

    def downstream_treatments(self, limit=None):
        from archaeological_finds.models_treatments import \
            FindDownstreamTreatments
        return self._get_treatments(FindDownstreamTreatments, 'downstream',
                                    limit=limit)

    def limited_downstream_treatments(self):
        return self.downstream_treatments(15)

    def all_treatments(self):
        return self.upstream_treatments() + self.downstream_treatments()

    def get_department(self):
        bf = self.get_first_base_find()
        if not bf:
            return "00"
        return bf.context_record.operation.get_department()

    def get_town_label(self):
        bf = self.get_first_base_find()
        if not bf:
            return "00"
        return bf.context_record.operation.get_town_label()

    @classmethod
    def get_periods(cls, slice='year', fltr={}):
        q = cls.objects
        if fltr:
            q = q.filter(**fltr)
        if slice == 'year':
            years = set()
            finds = q.filter(downstream_treatment__isnull=True)
            for find in finds:
                bi = find.base_finds.all()
                if not bi:
                    continue
                bi = bi[0]
                if bi.context_record.operation.start_date:
                    yr = bi.context_record.operation.start_date.year
                    years.add(yr)
        return list(years)

    @classmethod
    def get_by_year(cls, year, fltr={}):
        q = cls.objects
        if fltr:
            q = q.filter(**fltr)
        return q.filter(
            downstream_treatment__isnull=True,
            base_finds__context_record__operation__start_date__year=year)

    @classmethod
    def get_operations(cls):
        operations = set()
        finds = cls.objects.filter(downstream_treatment__isnull=True)
        for find in finds:
            bi = find.base_finds.all()
            if not bi:
                continue
            bi = bi[0]
            pk = bi.context_record.operation.pk
            operations.add(pk)
        return list(operations)

    @classmethod
    def get_by_operation(cls, operation_id):
        return cls.objects.filter(
            downstream_treatment__isnull=True,
            base_finds__context_record__operation__pk=operation_id)

    @classmethod
    def get_total_number(cls, fltr={}):
        q = cls.objects
        if fltr:
            q = q.filter(**fltr)
        return q.filter(downstream_treatment__isnull=True).count()

    def duplicate(self, user):
        model = self.__class__

        new = model.objects.get(pk=self.pk)

        for field in model._meta.fields:
            # pk is in PRIVATE_FIELDS so: new.pk = None and a new
            # item will be created on save
            if field.name in PRIVATE_FIELDS:
                setattr(new, field.name, None)
        new.order = self.order + 1
        new.history_order = user
        new.image.name = self.image.name
        # force_copy is necessary to not regenerate a thumb and resize
        # again the image
        new.save(force_copy=True)

        # m2m fields
        m2m = [field.name for field in model._meta.many_to_many
               if field.name not in PRIVATE_FIELDS]
        for field in m2m:
            for val in getattr(self, field).all():
                getattr(new, field).add(val)
        return new

    @classmethod
    def get_query_owns(cls, user):
        return (Q(base_finds__context_record__operation__scientist=user.
                 ishtaruser.person) |
                Q(base_finds__context_record__operation__in_charge=user.
                  ishtaruser.person) |
                Q(base_finds__context_record__operation__collaborators__pk=user.
                  ishtaruser.person.pk) |
                Q(history_creator=user)) \
           & Q(base_finds__context_record__operation__end_date__isnull=True)

    @classmethod
    def get_owns(cls, user, menu_filtr=None, limit=None,
                 values=None, get_short_menu_class=None):
        replace_query = None
        if menu_filtr and 'contextrecord' in menu_filtr:
            replace_query = Q(
                base_finds__context_record=menu_filtr['contextrecord']
            )
        owns = super(Find, cls).get_owns(
            user, replace_query=replace_query, limit=limit, values=values,
            get_short_menu_class=get_short_menu_class)
        return cls._return_get_owns(owns, values, get_short_menu_class)

    def _generate_cached_label(self):
        return unicode(self)

    @classmethod
    def cached_label_bulk_update(cls, operation_id=None, parcel_id=None,
                                 context_record_id=None):
        if operation_id:
            filters = """
            INNER JOIN find_first_base_find myfbf
                ON myfbf.find_id = myf.id
            INNER JOIN archaeological_finds_basefind mybf
                ON myfbf.basefind_id = mybf.id
            INNER JOIN archaeological_context_records_contextrecord acr
            ON acr.operation_id = %s AND acr.id = mybf.context_record_id
            """
            args = [int(operation_id)]
        elif parcel_id:
            filters = """
            INNER JOIN find_first_base_find myfbf
                ON myfbf.find_id = myf.id
            INNER JOIN archaeological_finds_basefind mybf
                ON myfbf.basefind_id = mybf.id
            INNER JOIN archaeological_context_records_contextrecord acr
            ON acr.parcel_id = %s AND acr.id = mybf.context_record_id
            """
            args = [int(parcel_id)]
        elif context_record_id:
            filters = """
            INNER JOIN find_first_base_find myfbf
                ON myfbf.find_id = myf.id
            INNER JOIN archaeological_finds_basefind mybf
                ON myfbf.basefind_id = mybf.id AND mybf.context_record_id = %s
            """
            args = [int(context_record_id)]
        else:
            return

        sql = """
        UPDATE "archaeological_finds_find" AS f
            SET cached_label =
                CASE
                WHEN find_cached_bulk_update.main_ope_code is NULL
                THEN
                    CASE
                    WHEN find_cached_bulk_update.year
                            IS NOT NULL
                        AND find_cached_bulk_update.ope_code
                            IS NOT NULL
                    THEN
                       '{ope_prefix}' ||
                       find_cached_bulk_update.year ||
                       '-' ||
                       find_cached_bulk_update.ope_code
                    ELSE ''
                    END
                ELSE
                    '{main_ope_prefix}' ||
                    find_cached_bulk_update.main_ope_code
                END
                || '-' ||
                to_char(find_cached_bulk_update.index, 'fm{zeros}')
                || '{join}' ||
                find_cached_bulk_update.label


            FROM find_cached_bulk_update
            WHERE f.id = find_cached_bulk_update.id
                    AND f.id IN (
                SELECT myf.id FROM archaeological_finds_find myf
                {filters}
            );
        """.format(main_ope_prefix=settings.ISHTAR_OPE_PREFIX,
                   ope_prefix=settings.ISHTAR_DEF_OPE_PREFIX,
                   join=settings.JOINT, filters=filters,
                   zeros=settings.ISHTAR_FINDS_INDEX_ZERO_LEN * "0")
        # with connection.cursor() as c:  # django 1.8
        c = connection.cursor()
        c.execute(sql, args)
        transaction.commit_unless_managed()

    def save(self, *args, **kwargs):
        super(Find, self).save(*args, **kwargs)

        updated = False
        self.skip_history_when_saving = True
        if not self.external_id or self.auto_external_id:
            external_id = get_external_id('find_external_id', self)
            if external_id != self.external_id:
                updated = True
                self.auto_external_id = True
                self.external_id = external_id
        if updated:
            self._cached_label_checked = False
            self.save()
            return

        q = self.base_finds
        if not self.index and q.count():
            operation = q.filter(
                context_record__operation__pk__isnull=False).order_by(
                '-context_record__operation__start_date')
            if operation.count():
                operation = operation.all()[0].context_record.operation
                q = Find.objects \
                    .filter(base_finds__context_record__operation=operation)
                if self.pk:
                    q = q.exclude(pk=self.pk)
                if q.count():
                    self.index = q.aggregate(Max('index'))['index__max'] + 1
                else:
                    self.index = 1
                self._cached_label_checked = False
                self.save()
        for base_find in self.base_finds.filter(
                context_record__operation__pk__isnull=False).all():
            modified = False
            if not base_find.index:
                modified = True
                base_find.index = BaseFind.get_max_index(
                    base_find.context_record.operation) + 1
            short_id = base_find.short_id()
            if base_find.cache_short_id != short_id:
                base_find.cache_short_id = short_id
                modified = True
            complete_id = base_find.complete_id()
            if base_find.cache_complete_id != complete_id:
                base_find.cache_complete_id = complete_id
                modified = True
            if modified:
                base_find.skip_history_when_saving = True
                base_find._cached_label_checked = False
                base_find.save()
                # if not base_find.material_index:
                #    idx = BaseFind.objects\
                #                  .filter(context_record=base_find.context_record,
                #                          find__material_types=self.material_type)\
                #                  .aggregate(Max('material_index'))
                #    base_find.material_index = \
                #        idx and idx['material_index__max'] + 1 or 1


post_save.connect(cached_label_changed, sender=Find)


def base_find_find_changed(sender, **kwargs):
    obj = kwargs.get('instance', None)
    if not obj:
        return
    # recalculate complete id and external id
    obj.save()

m2m_changed.connect(base_find_find_changed, sender=Find.base_finds.through)


class FindSource(Source):
    SHOW_URL = 'show-findsource'
    MODIFY_URL = 'find_source_modify'
    TABLE_COLS = [
        'find__base_finds__context_record__operation',
        'find__base_finds__context_record', 'find'] + \
        Source.TABLE_COLS

    # search parameters
    BOOL_FIELDS = ['duplicate']
    RELATIVE_SESSION_NAMES = [
        ('find', 'find__pk'),
        ('contextrecord', 'find__base_finds__context_record__pk'),
        ('operation', 'find__base_finds__context_record__operation__pk'),
        ('file',
         'find__base_finds__context_record__operation__associated_file__pk')
    ]
    EXTRA_REQUEST_KEYS = {
        'title': 'title__icontains',
        'description': 'description__icontains',
        'comment': 'comment__icontains',
        'additional_information': 'additional_information__icontains',
        'person': 'authors__person__pk',
        'find__base_finds__context_record__operation__year':
            'find__base_finds__context_record__operation__year',
        'find__base_finds__context_record__operation__operation_code':
            'find__base_finds__context_record__operation__operation_code',
        'find__base_finds__context_record__operation__code_patriarche':
            'find__base_finds__context_record__operation__code_patriarche',
        'find__datings__period': 'find__datings__period__pk',
        'find__description': 'find__description__icontains',
    }

    class Meta:
        verbose_name = _(u"Find documentation")
        verbose_name_plural = _(u"Find documentations")
        permissions = (
            ("view_findsource",
             ugettext(u"Can view all Find sources")),
            ("view_own_findsource",
             ugettext(u"Can view own Find source")),
            ("add_own_findsource",
             ugettext(u"Can add own Find source")),
            ("change_own_findsource",
             ugettext(u"Can change own Find source")),
            ("delete_own_findsource",
             ugettext(u"Can delete own Find source")),
        )
    find = models.ForeignKey(Find, verbose_name=_(u"Find"),
                             related_name="source")

    @property
    def owner(self):
        return self.find

    @classmethod
    def get_query_owns(cls, user):
        return (Q(find__base_finds__context_record__operation__scientist=user.
                  ishtaruser.person) |
                Q(find__base_finds__context_record__operation__in_charge=user.
                  ishtaruser.person) |
                Q(
            find__base_finds__context_record__operation__collaborators__pk=user.
            ishtaruser.person.pk)) \
               & Q(
            find__base_finds__context_record__operation__end_date__isnull=True)


class Property(LightHistorizedItem):
    find = models.ForeignKey(Find, verbose_name=_(u"Find"))
    administrative_act = models.ForeignKey(
        AdministrativeAct, verbose_name=_(u"Administrative act"))
    person = models.ForeignKey(Person, verbose_name=_(u"Person"),
                               related_name='properties')
    start_date = models.DateField(_(u"Start date"))
    end_date = models.DateField(_(u"End date"))

    class Meta:
        verbose_name = _(u"Property")
        verbose_name_plural = _(u"Properties")

    def __unicode__(self):
        return self.person + settings.JOINT + self.find



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010-2016 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

from django.conf.urls import *

from ishtar_common.wizards import check_rights
import views

from archaeological_finds import models

# be careful: each check_rights must be relevant with ishtar_menu

# forms
urlpatterns = patterns(
    '',
    url(r'find_search/(?P<step>.+)?$',
        check_rights(['view_find', 'view_own_find'])(
            views.find_search_wizard), name='find_search'),
    url(r'find_creation/(?P<step>.+)?$',
        check_rights(['add_find', 'add_own_find'])(
            views.find_creation_wizard), name='find_creation'),
    url(r'find_modification/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.find_modification_wizard), name='find_modification'),
    url(r'find_deletion/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.find_deletion_wizard), name='find_deletion'),
    url(r'find_modify/(?P<pk>.+)/$',
        views.find_modify, name='find_modify'),
    url(r'find_source_search/(?P<step>.+)?$',
        check_rights(['view_find', 'view_own_find'])(
            views.find_source_search_wizard),
        name='find_source_search'),
    url(r'find_source_creation/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.find_source_creation_wizard),
        name='find_source_creation'),
    url(r'find_source_modification/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.find_source_modification_wizard),
        name='find_source_modification'),
    url(r'find_source_modify/(?P<pk>.+)/$',
        views.find_source_modify, name='find_source_modify'),
    url(r'find_source_deletion/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.find_source_deletion_wizard),
        name='find_source_deletion'),
    url(r'^find_basket_creation/$',
        check_rights(['change_find', 'change_own_find'])(
            views.NewFindBasketView.as_view()), name='new_findbasket'),
    url(r'^find_basket_modification_add/$',
        check_rights(['change_find', 'change_own_find'])(
            views.SelectBasketForManagement.as_view()),
        name='select_findbasketforadd'),
    url(r'^find_basket_modification_add/(?P<pk>[0-9]+)?/$',
        check_rights(['change_find', 'change_own_find'])(
            views.SelectItemsInBasket.as_view()),
        name='select_itemsinbasket'),
    url(r'^find_basket_modification_add_item/$',
        check_rights(['change_find', 'change_own_find'])(
            views.FindBasketAddItemView.as_view()),
        name='add_iteminbasket'),
    url(r'^find_basket_modification_delete_item/(?P<basket>[0-9]+)?'
        r'/(?P<find_pk>[0-9]+)?/$',
        check_rights(['change_find', 'change_own_find'])(
            views.FindBasketDeleteItemView.as_view()),
        name='delete_iteminbasket'),
    url(r'^find_basket_list/(?P<pk>[0-9]+)?/$',
        check_rights(['change_find', 'change_own_find'])(
            views.FindBasketListView.as_view()),
        name='list_iteminbasket'),
    url(r'^find_basket_deletion/$',
        check_rights(['change_find', 'change_own_find'])(
            views.DeleteFindBasketView.as_view()), name='delete_findbasket'),

    url(r'^treatment_creation/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.treatment_creation_wizard), name='treatment_creation'),
    url(r'^treatment_modification/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.treatment_modification_wizard),
        name='treatment_modification'),
    url(r'treatment_modify/(?P<pk>.+)/$',
        views.treatment_modify, name='treatment_modify'),
    url(r'^treatment_search/(?P<step>.+)?$',
        check_rights(['view_find', 'view_own_find'])(
            views.treatment_search_wizard), name='treatment_search'),
    url(r'^treatment_deletion/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.treatment_deletion_wizard), name='treatment_deletion'),

    url(r'^show-treatmentsource(?:/(?P<pk>.+))?/(?P<type>.+)?$',
        views.show_treatmentsource,
        name=models.TreatmentSource.SHOW_URL),
    url(r'^get-treatmentsource/(?P<type>.+)?$',
        views.get_treatmentsource,
        name='get-treatmentsource'),
    url(r'^treatment_source_search/(?P<step>.+)?$',
        check_rights(['view_treatment', 'view_own_treatment'])(
            views.treatment_source_search_wizard),
        name='treatment_source_search'),
    url(r'^treatment_source_creation/(?P<step>.+)?$',
        check_rights(['change_treatment', 'change_own_treatment'])(
            views.treatment_source_creation_wizard),
        name='treatment_source_creation'),
    url(r'^treatment_source_modification/(?P<step>.+)?$',
        check_rights(['change_treatment', 'change_own_treatment'])(
            views.treatment_source_modification_wizard),
        name='treatment_source_modification'),
    url(r'^treatment_source_modify/(?P<pk>.+)/$',
        views.treatment_source_modify, name='treatment_source_modify'),
    url(r'^treatment_source_deletion/(?P<step>.+)?$',
        check_rights(['change_treatment', 'change_own_treatment'])(
            views.treatment_source_deletion_wizard),
        name='treatment_source_deletion'),

    url(r'^treatment_admacttreatment_search/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatment_administrativeact_search_wizard),
        name='treatment_admacttreatment_search'),
    url(r'^treatment_admacttreatment/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatment_administrativeact_wizard),
        name='treatment_admacttreatment'),
    url(r'^treatment_admacttreatment_modification/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatment_administrativeact_modification_wizard),
        name='treatment_admacttreatment_modification'),
    url(r'^treatment_administrativeacttreatment_modify/(?P<pk>.+)/$',
        views.treatment_administrativeacttreatment_modify,
        name='treatment_administrativeacttreatment_modify'),
    url(r'^treatment_admacttreatment_deletion/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatment_admacttreatment_deletion_wizard),
        name='treatment_admacttreatment_deletion'),
    url(r'^get-administrativeacttreatment/(?P<type>.+)?$',
        views.get_administrativeacttreatment,
        name='get-administrativeacttreatment'),

    url(r'^treatmentfle_admacttreatmentfle_search/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatmentfile_admacttreatmentfile_search_wizard),
        name='treatmentfle_admacttreatmentfle_search'),
    url(r'^treatmentfle_admacttreatmentfle_modification/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatmentfile_admacttreatmentfile_modification_wizard),
        name='treatmentfle_admacttreatmentfle_modification'),
    url(r'^treatmentfle_admacttreatmentfle/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatmentfile_admacttreatmentfile_wizard),
        name='treatmentfle_admacttreatmentfle'),
    url(r'^treatmentfile_administrativeacttreatmentfile_modify/(?P<pk>.+)/$',
        views.treatmentfile_administrativeacttreatmentfile_modify,
        name='treatmentfile_administrativeacttreatmentfile_modify'),
    url(r'^treatmentfle_admacttreatmentfle_deletion/(?P<step>.+)?$',
        check_rights(['change_administrativeact'])(
            views.treatmentfile_admacttreatmentfile_deletion_wizard),
        name='treatmentfle_admacttreatmentfle_deletion'),

    url(r'^show-treatmentfilesource(?:/(?P<pk>.+))?/(?P<type>.+)?$',
        views.show_treatmentfilesource,
        name=models.TreatmentFileSource.SHOW_URL),
    url(r'^get-treatmentfilesource/(?P<type>.+)?$',
        views.get_treatmentfilesource,
        name='get-treatmentfilesource'),
    url(r'^treatmentfile_source_search/(?P<step>.+)?$',
        check_rights(['view_treatmentfile', 'view_own_treatmentfile'])(
            views.treatmentfile_source_search_wizard),
        name='treatmentfile_source_search'),
    url(r'^treatmentfile_source_creation/(?P<step>.+)?$',
        check_rights(['change_treatmentfile', 'change_own_treatmentfile'])(
            views.treatmentfile_source_creation_wizard),
        name='treatmentfile_source_creation'),
    url(r'^treatmentfile_source_modification/(?P<step>.+)?$',
        check_rights(['change_treatmentfile', 'change_own_treatmentfile'])(
            views.treatmentfile_source_modification_wizard),
        name='treatmentfile_source_modification'),
    url(r'^treatmentfile_source_modify/(?P<pk>.+)/$',
        views.treatmentfile_source_modify, name='treatmentfile_source_modify'),
    url(r'^treatmentfile_source_deletion/(?P<step>.+)?$',
        check_rights(['change_treatmentfile', 'change_own_treatmentfile'])(
            views.treatmentfile_source_deletion_wizard),
        name='treatmentfile_source_deletion'),


    url(r'^treatmentfle_search/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.treatmentfile_search_wizard),
        name='treatmentfile_search'),
    url(r'treatmentfle_creation/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.treatmentfile_creation_wizard),
        name='treatmentfile_creation'),
    url(r'treatmentfle_modification/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.treatmentfile_modification_wizard),
        name='treatmentfile_modification'),
    url(r'^treatmentfile_modify/(?P<pk>.+)/$',
        views.treatmentfile_modify, name='treatmentfile_modify'),
    url(r'^treatmentfle_deletion/(?P<step>.+)?$',
        check_rights(['change_find', 'change_own_find'])(
            views.treatmentfile_deletion_wizard),
        name='treatmentfile_deletion'),
    url(r'get-administrativeacttreatmentfile/(?P<type>.+)?$',
        views.get_administrativeacttreatmentfile,
        name='get-administrativeacttreatmentfile'),
    url(r'get-upstreamtreatment/(?P<type>.+)?$', views.get_upstreamtreatment,
        name='get-upstreamtreatment'),
    url(r'get-downstreamtreatment/(?P<type>.+)?$',
        views.get_downstreamtreatment,
        name='get-downstreamtreatment'),
)

urlpatterns += patterns(
    'archaeological_finds.views',
    url(r'autocomplete-objecttype/$', 'autocomplete_objecttype',
        name='autocomplete-objecttype'),
    url(r'autocomplete-materialtype/$', 'autocomplete_materialtype',
        name='autocomplete-materialtype'),
    url(r'autocomplete-preservationtype/$', 'autocomplete_preservationtype',
        name='autocomplete-preservationtype'),
    url(r'autocomplete-integritytype/$', 'autocomplete_integritytype',
        name='autocomplete-integritytype'),
    url(r'autocomplete-treatmentfile/$', 'autocomplete_treatmentfile',
        name='autocomplete-treatmentfile'),
    url(r'get-find/own/(?P<type>.+)?$', 'get_find',
        name='get-own-find', kwargs={'force_own': True}),
    url(r'get-find/(?P<type>.+)?$', 'get_find',
        name='get-find'),
    url(r'get-find-for-ope/own/(?P<type>.+)?$', 'get_find_for_ope',
        name='get-own-find-for-ope', kwargs={'force_own': True}),
    url(r'get-find-for-ope/(?P<type>.+)?$', 'get_find_for_ope',
        name='get-find-for-ope'),
    url(r'get-find-for-treatment/own/(?P<type>.+)?$', 'get_find_for_treatment',
        name='get-own-find-for-treatment', kwargs={'force_own': True}),
    url(r'get-find-for-treatment/(?P<type>.+)?$', 'get_find_for_treatment',
        name='get-find-for-treatment'),
    url(r'get-find-full/own/(?P<type>.+)?$', 'get_find',
        name='get-own-find-full', kwargs={'full': True, 'force_own': True}),
    url(r'get-find-full/(?P<type>.+)?$', 'get_find',
        name='get-find-full', kwargs={'full': True}),
    url(r'get-find-shortcut/(?P<type>.+)?$',
        'get_find', name='get-find-shortcut',
        kwargs={'full': 'shortcut'}),
    url(r'get-findsource/(?P<type>.+)?$',
        'get_findsource', name='get-findsource'),
    url(r'get-findsource-full/(?P<type>.+)?$',
        'get_findsource', name='get-findsource-full',
        kwargs={'full': True}),
    url(r'^show-findsource(?:/(?P<pk>.+))?/(?P<type>.+)?$', 'show_findsource',
        name=models.FindSource.SHOW_URL),
    url(r'^show-find/basket-(?P<pk>.+)/(?P<type>.+)?$', 'show_findbasket',
        name='show-findbasket'),
    url(r'^display-find/basket-(?P<pk>.+)/$', 'display_findbasket',
        name='display-findbasket'),
    url(r'^show-find(?:/(?P<pk>.+))?/(?P<type>.+)?$', 'show_find',
        name=models.Find.SHOW_URL),
    url(r'^display-find/(?P<pk>.+)/$', 'display_find',
        name='display-' + models.Find.SLUG),
    url(r'^show-historized-find/(?P<pk>.+)?/(?P<date>.+)?$',
        'show_find', name='show-historized-find'),
    url(r'^revert-find/(?P<pk>.+)/(?P<date>.+)$',
        'revert_find', name='revert-find'),
    url(r'^get-treatment/(?P<type>.+)?$',
        'get_treatment', name='get-treatment'),
    url(r'get-treatment-shortcut/(?P<type>.+)?$',
        'get_treatment', name='get-treatment-shortcut',
        kwargs={'full': 'shortcut'}),
    url(r'^show-treatment(?:/(?P<pk>.+))?/(?P<type>.+)?$', 'show_treatment',
        name=models.Treatment.SHOW_URL),
    url(r'show-historized-treatment/(?P<pk>.+)?/(?P<date>.+)?$',
        'show_treatment', name='show-historized-treatment'),
    url(r'^revert-treatment/(?P<pk>.+)/(?P<date>.+)$',
        'revert_treatment', name='revert-treatment'),
    url(r'get-treatmentfile/(?P<type>.+)?$',
        'get_treatmentfile', name='get-treatmentfile'),
    url(r'get-treatmentfile-shortcut/(?P<type>.+)?$',
        'get_treatmentfile', name='get-treatmentfile-shortcut',
        kwargs={'full': 'shortcut'}),
    url(r'^show-treatmentfile(?:/(?P<pk>.+))?/(?P<type>.+)?$',
        'show_treatmentfile',
        name=models.TreatmentFile.SHOW_URL),
    url(r'show-historized-treatmentfile/(?P<pk>.+)?/(?P<date>.+)?$',
        'show_treatmentfile', name='show-historized-treatmentfile'),
    url(r'^revert-treatmentfile/(?P<pk>.+)/(?P<date>.+)$',
        'revert_treatmentfile', name='revert-treatmentfile'),
    # url(r'show-treatmentfile(?:/(?P<pk>.+))?/(?P<type>.+)?$',
    # 'show_treatmentfile',
    #     name=models.TreatmentFile.SHOW_URL),
)

urlpatterns += patterns(
    'archaeological_operations.views',
    url(r'^treatment_administrativeact_document/$',
        'administrativeactfile_document',
        name='treatment-administrativeact-document',
        kwargs={'treatment': True}),
    url(r'^treatmentfle_administrativeact_document/$',
        'administrativeactfile_document',
        name='treatmentfle-administrativeact-document',
        kwargs={'treatment_file': True}),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010-2016  Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

from django import forms
from django.conf import settings
from django.forms.formsets import formset_factory
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from ishtar_common.models import Person, valid_id
from archaeological_finds.models import TreatmentType, FindBasket
import models
from ishtar_common import widgets
from ishtar_common.forms import name_validator, reverse_lazy, \
    get_form_selection, TableSelect, ManageOldType, FinalForm, FormSet
from archaeological_finds.forms import FindMultipleFormSelection, \
    SelectFindBasketForm


def get_warehouse_field(label=_(u"Warehouse"), required=True):
    # !FIXME hard_link, reverse_lazy doen't seem to work with formsets
    url = "/" + settings.URL_PATH + 'autocomplete-warehouse'
    widget = widgets.JQueryAutoComplete(url, associated_model=models.Warehouse)
    return forms.IntegerField(widget=widget, label=label, required=required,
                              validators=[valid_id(models.Warehouse)])


class SelectedDivisionForm(ManageOldType, forms.Form):
    form_label = _(u"Division")
    base_model = 'associated_division'
    associated_models = {'division': models.WarehouseDivision,
                         'associated_division': models.WarehouseDivisionLink}
    division = forms.ChoiceField(
        label=_(u"Division"), choices=(), required=False,
        validators=[valid_id(models.WarehouseDivision)])
    order = forms.IntegerField(_(u"Order"), initial=10, required=False)

    def __init__(self, *args, **kwargs):
        super(SelectedDivisionForm, self).__init__(*args, **kwargs)
        self.fields['division'].choices = \
            models.WarehouseDivision.get_types(
            initial=self.init_data.get('division')
        )


class DivisionFormSet(FormSet):
    def clean(self):
        """Checks that no divisions are duplicated."""
        return self.check_duplicate(('division',),
                                    _("There are identical divisions."))

SelectedDivisionFormset = formset_factory(
    SelectedDivisionForm, can_delete=True, formset=DivisionFormSet)
SelectedDivisionFormset.form_label = _(u"Divisions")


class WarehouseSelect(TableSelect):
    name = forms.CharField(label=_(u"Name"))
    warehouse_type = forms.ChoiceField(label=_(u"Warehouse type"), choices=[])
    towns = forms.CharField(label=_(u"Town"))

    def __init__(self, *args, **kwargs):
        super(WarehouseSelect, self).__init__(*args, **kwargs)
        self.fields['warehouse_type'].choices = \
            models.WarehouseType.get_types()
        self.fields['warehouse_type'].help_text = \
            models.WarehouseType.get_help()


class WarehouseFormSelection(forms.Form):
    form_label = _("Warehouse search")
    associated_models = {'pk': models.Warehouse}
    currents = {'pk': models.Warehouse}
    pk = forms.IntegerField(
        label="", required=False,
        widget=widgets.JQueryJqGrid(
            reverse_lazy('get-warehouse'),
            WarehouseSelect, models.Warehouse),
        validators=[valid_id(models.Warehouse)])


class WarehouseForm(ManageOldType, forms.Form):
    form_label = _(u"Warehouse")
    associated_models = {'warehouse_type': models.WarehouseType,
                         'person_in_charge': Person}

    name = forms.CharField(label=_(u"Name"), max_length=200,
                           validators=[name_validator])
    warehouse_type = forms.ChoiceField(label=_(u"Warehouse type"),
                                       choices=[])
    person_in_charge = forms.IntegerField(
        label=_(u"Person in charge"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person'),
            associated_model=Person, new=True),
        validators=[valid_id(Person)],
        required=False)
    comment = forms.CharField(label=_(u"Comment"), widget=forms.Textarea,
                              required=False)
    address = forms.CharField(label=_(u"Address"), widget=forms.Textarea,
                              required=False)
    address_complement = forms.CharField(label=_(u"Address complement"),
                                         widget=forms.Textarea, required=False)
    postal_code = forms.CharField(label=_(u"Postal code"), max_length=10,
                                  required=False)
    town = forms.CharField(label=_(u"Town"), max_length=30, required=False)
    country = forms.CharField(label=_(u"Country"), max_length=30,
                              required=False)
    phone = forms.CharField(label=_(u"Phone"), max_length=18, required=False)
    mobile_phone = forms.CharField(label=_(u"Mobile phone"), max_length=18,
                                   required=False)

    def __init__(self, *args, **kwargs):
        if 'limits' in kwargs:
            kwargs.pop('limits')
        super(WarehouseForm, self).__init__(*args, **kwargs)
        self.fields['warehouse_type'].choices = \
            models.WarehouseType.get_types(
                initial=self.init_data.get('warehouse_type'))
        self.fields['warehouse_type'].help_text = \
            models.WarehouseType.get_help()

    def save(self, user):
        dct = self.cleaned_data
        dct['history_modifier'] = user
        dct['warehouse_type'] = models.WarehouseType.objects.get(
            pk=dct['warehouse_type'])
        if 'person_in_charge' in dct and dct['person_in_charge']:
            dct['person_in_charge'] = models.Person.objects.get(
                pk=dct['person_in_charge'])
        new_item = models.Warehouse(**dct)
        new_item.save()
        return new_item


class WarehouseDeletionForm(FinalForm):
    confirm_msg = _(u"Would you like to delete this warehouse?")
    confirm_end_msg = _(u"Would you like to delete this warehouse?")


class ContainerForm(ManageOldType, forms.Form):
    form_label = _(u"Container")
    file_upload = True
    associated_models = {'container_type': models.ContainerType,
                         'location': models.Warehouse,
                         'responsible': models.Warehouse}
    reference = forms.CharField(label=_(u"Ref."))
    container_type = forms.ChoiceField(label=_(u"Container type"), choices=[])
    location = forms.IntegerField(
        label=_(u"Current location (warehouse)"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-warehouse'),
            associated_model=models.Warehouse, new=True),
        validators=[valid_id(models.Warehouse)])
    responsible = forms.IntegerField(
        label=_(u"Responsible warehouse"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-warehouse'),
            associated_model=models.Warehouse, new=True),
        validators=[valid_id(models.Warehouse)])
    image = forms.ImageField(
        label=_(u"Image"), help_text=mark_safe(
            _(u"<p>Heavy images are resized to: %(width)dx%(height)d "
              u"(ratio is preserved).</p>") % {
                'width': settings.IMAGE_MAX_SIZE[0],
                'height': settings.IMAGE_MAX_SIZE[1]}),
        max_length=255, required=False, widget=widgets.ImageFileInput())
    comment = forms.CharField(label=_(u"Comment"),
                              widget=forms.Textarea, required=False)

    def __init__(self, *args, **kwargs):
        if 'limits' in kwargs:
            kwargs.pop('limits')
        super(ContainerForm, self).__init__(*args, **kwargs)
        self.fields['container_type'].choices = \
            models.ContainerType.get_types(
                initial=self.init_data.get('container_type'))
        self.fields['container_type'].help_text = \
            models.ContainerType.get_help()

    def save(self, user):
        dct = self.cleaned_data
        dct['history_modifier'] = user
        dct['container_type'] = models.ContainerType.objects.get(
            pk=dct['container_type'])
        dct['location'] = models.Warehouse.objects.get(pk=dct['location'])
        dct['responsible'] = models.Warehouse.objects.get(pk=dct['responsible'])
        new_item = models.Container(**dct)
        new_item.save()
        return new_item


class ContainerModifyForm(ContainerForm):
    pk = forms.IntegerField(required=False, widget=forms.HiddenInput)
    index = forms.IntegerField(label=_(u"ID"))

    def __init__(self, *args, **kwargs):
        super(ContainerModifyForm, self).__init__(*args, **kwargs)
        self.fields.keyOrder.pop(self.fields.keyOrder.index('index'))
        self.fields.keyOrder.insert(
            self.fields.keyOrder.index("location") + 1, 'index')

    def clean(self):
        # manage unique ID
        cleaned_data = self.cleaned_data
        index = cleaned_data.get("index")
        warehouse = cleaned_data.get("location")
        q = models.Container.objects.filter(
            index=index, location__pk=warehouse)
        if 'pk' in cleaned_data and cleaned_data['pk']:
            q = q.exclude(pk=int(cleaned_data['pk']))
        if q.count():
            raise forms.ValidationError(_(u"This ID already exists for "
                                          u"this warehouse."))
        return cleaned_data


class ContainerSelect(TableSelect):
    location = get_warehouse_field()
    container_type = forms.ChoiceField(label=_(u"Container type"), choices=[])
    reference = forms.CharField(label=_(u"Ref."))

    def __init__(self, *args, **kwargs):
        super(ContainerSelect, self).__init__(*args, **kwargs)
        self.fields['container_type'].choices = \
            models.ContainerType.get_types()
        self.fields['container_type'].help_text = \
            models.ContainerType.get_help()

ContainerFormSelection = get_form_selection(
    'ContainerFormSelection', _(u"Container search"), 'container',
    models.Container, ContainerSelect, 'get-container',
    _(u"You should select a container."), new=True,
    new_message=_(u"Add a new container"))

MainContainerFormSelection = get_form_selection(
    'ContainerFormSelection', _(u"Container search"), 'pk',
    models.Container, ContainerSelect, 'get-container',
    _(u"You should select a container.")
)


class BasePackagingForm(SelectFindBasketForm):
    form_label = _(u"Packaging")
    associated_models = {'treatment_type': TreatmentType,
                         'person': Person,
                         'location': models.Warehouse,
                         'basket': FindBasket}
    person = forms.IntegerField(
        label=_(u"Packager"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person'),
            associated_model=Person, new=True),
        validators=[valid_id(Person)])
    start_date = forms.DateField(
        label=_(u"Date"), required=False, widget=widgets.JQueryDate)


class FindPackagingFormSelection(FindMultipleFormSelection):
    form_label = _(u"Packaged finds")


class LocalisationForm(forms.Form):
    form_label = _(u"Localisation")

    def __init__(self, *args, **kwargs):
        self.container, self.warehouse = None, None
        if 'warehouse' in kwargs:
            self.warehouse = kwargs.pop('warehouse')
        if 'container' in kwargs:
            self.container = kwargs.pop('container')
        super(LocalisationForm, self).__init__(*args, **kwargs)
        if not self.warehouse:
            return
        for divlink in self.warehouse.warehousedivisionlink_set.order_by(
                'order').all():
            initial = u"-"
            if self.container:
                q = models.ContainerLocalisation.objects.filter(
                    division=divlink, container=self.container)
                if q.count():
                    initial = q.all()[0].reference
            self.fields['division_{}'.format(divlink.pk)] = forms.CharField(
                label=str(divlink.division), max_length=200, initial=initial,
            )


class ContainerDeletionForm(FinalForm):
    confirm_msg = _(u"Would you like to delete this container?")
    confirm_end_msg = _(u"Would you like to delete this container?")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2017 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

import datetime

from django.conf import settings
from django.contrib.gis.db import models
from django.core.cache import cache
from django.db.models import Q, Count, Sum
from django.db.models.signals import post_save, m2m_changed, post_delete
from django.utils.translation import ugettext_lazy as _, ugettext

from ishtar_common.utils import cached_label_changed, get_cache

from ishtar_common.models import GeneralType, BaseHistorizedItem, \
    HistoricalRecords, OwnPerms, Person, Organization, Department, Town, \
    Dashboard, DashboardFormItem, ValueGetter, ShortMenuItem, \
    OperationType, get_external_id, post_save_cache

from archaeological_operations.models import get_values_town_related, \
    ClosedItem


class FileType(GeneralType):
    class Meta:
        verbose_name = _(u"Archaeological file type")
        verbose_name_plural = _(u"Archaeological file types")
        ordering = ('label',)

    @classmethod
    def is_preventive(cls, file_type_id, key=''):
        key = key or 'preventive'
        try:
            preventive = FileType.get_cache(key).pk
            return file_type_id == preventive
        except FileType.DoesNotExist:
            return False
post_save.connect(post_save_cache, sender=FileType)
post_delete.connect(post_save_cache, sender=FileType)


class PermitType(GeneralType):
    class Meta:
        verbose_name = _(u"Permit type")
        verbose_name_plural = _(u"Permit types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=PermitType)
post_delete.connect(post_save_cache, sender=PermitType)

if settings.COUNTRY == 'fr':
    class SaisineType(GeneralType, ValueGetter):
        delay = models.IntegerField(_(u"Delay (in days)"), default=30)

        class Meta:
            verbose_name = u"Type de saisine"
            verbose_name_plural = u"Types de saisine"
            ordering = ('label',)
    post_save.connect(post_save_cache, sender=SaisineType)
    post_delete.connect(post_save_cache, sender=SaisineType)


class File(ClosedItem, BaseHistorizedItem, OwnPerms, ValueGetter,
           ShortMenuItem, DashboardFormItem):
    SLUG = 'file'
    SHOW_URL = 'show-file'
    TABLE_COLS = ['numeric_reference', 'year', 'internal_reference',
                  'file_type', 'saisine_type', 'towns', ]

    # search parameters
    BOOL_FIELDS = ['end_date__isnull']
    EXTRA_REQUEST_KEYS = {
        'parcel_0': ('parcels__section',
                     'operations__parcels__section'),
        'parcel_1': ('parcels__parcel_number',
                     'operations__parcels__parcel_number'),
        'parcel_2': ('operations__parcels__public_domain',
                     'parcels__public_domain'),
        'end_date': 'end_date__isnull',
        'towns__numero_insee__startswith':
        'towns__numero_insee__startswith',
        'name': 'name__icontains',
        'cached_label': 'cached_label__icontains',
        'comment': 'comment__icontains',
        'permit_reference': 'permit_reference__icontains',
        'general_contractor__attached_to':
            'general_contractor__attached_to__pk',
        'history_creator': 'history_creator__ishtaruser__person__pk',
        'history_modifier': 'history_modifier__ishtaruser__person__pk',
    }

    # fields
    year = models.IntegerField(_(u"Year"),
                               default=lambda: datetime.datetime.now().year)
    numeric_reference = models.IntegerField(
        _(u"Numeric reference"), blank=True, null=True)
    internal_reference = models.CharField(_(u"Internal reference"), blank=True,
                                          null=True, max_length=60)
    external_id = models.CharField(_(u"External ID"), blank=True, null=True,
                                   max_length=120)
    auto_external_id = models.BooleanField(
        _(u"External ID is set automatically"), default=False)
    name = models.TextField(_(u"Name"), blank=True, null=True)
    file_type = models.ForeignKey(FileType, verbose_name=_(u"File type"))
    in_charge = models.ForeignKey(Person, related_name='file_responsability',
                                  verbose_name=_(u"Person in charge"),
                                  on_delete=models.SET_NULL,
                                  blank=True, null=True)
    general_contractor = models.ForeignKey(
        Person, related_name='general_contractor_files',
        verbose_name=_(u"General contractor"), blank=True, null=True,
        on_delete=models.SET_NULL,)  # aménageur - personne
    raw_general_contractor = models.CharField(_(
        u"General contractor (raw)"), max_length=200, blank=True, null=True)
    corporation_general_contractor = models.ForeignKey(
        Organization,
        related_name='general_contractor_files',
        verbose_name=_(u"General contractor organization"), blank=True,
        null=True, on_delete=models.SET_NULL,)  # aménageur
    responsible_town_planning_service = models.ForeignKey(
        Person, related_name='responsible_town_planning_service_files',
        blank=True, null=True,
        verbose_name=_(u"Responsible for planning service"),
        on_delete=models.SET_NULL,)  # service instructeur - personne
    raw_town_planning_service = models.CharField(
        _(u"Planning service (raw)"), max_length=200,
        blank=True, null=True)
    planning_service = models.ForeignKey(
        Organization,
        related_name='planning_service_files',
        blank=True, null=True,
        verbose_name=_(u"Planning service organization"),
        on_delete=models.SET_NULL,)  # service instructeur
    permit_type = models.ForeignKey(
        PermitType, verbose_name=_(u"Permit type"), blank=True, null=True)
    permit_reference = models.TextField(_(u"Permit reference"), blank=True,
                                        null=True)
    end_date = models.DateField(_(u"Closing date"), null=True, blank=True)
    main_town = models.ForeignKey(Town, verbose_name=_(u"Town"), null=True,
                                  blank=True, related_name='file_main')
    towns = models.ManyToManyField(Town, verbose_name=_(u"Towns"),
                                   related_name='file')
    creation_date = models.DateField(
        _(u"Creation date"), default=datetime.date.today, blank=True,
        null=True)
    reception_date = models.DateField(_(u'Reception date'), blank=True,
                                      null=True)
    related_file = models.ForeignKey("File", verbose_name=_(u"Related file"),
                                     blank=True, null=True)
    if settings.COUNTRY == 'fr':
        saisine_type = models.ForeignKey(
            SaisineType, blank=True, null=True,
            verbose_name=u"Type de saisine")
        instruction_deadline = models.DateField(_(u'Instruction deadline'),
                                                blank=True, null=True)
    total_surface = models.FloatField(_(u"Total surface (m2)"),
                                      blank=True, null=True)
    total_developed_surface = models.FloatField(
        _(u"Total developed surface (m2)"), blank=True, null=True)
    locality = models.CharField(_(u"Locality"),
                                max_length=100, null=True, blank=True)
    address = models.TextField(_(u"Main address"), null=True, blank=True)
    postal_code = models.CharField(_(u"Main address - postal code"),
                                   max_length=10, null=True, blank=True)
    comment = models.TextField(_(u"Comment"), null=True, blank=True)
    # research archaeology -->
    departments = models.ManyToManyField(
        Department, verbose_name=_(u"Departments"), null=True, blank=True)
    requested_operation_type = models.ForeignKey(
        OperationType, related_name='+', null=True,
        blank=True, verbose_name=_(u"Requested operation type"))
    organization = models.ForeignKey(
        Organization, blank=True, null=True, verbose_name=_(u"Organization"),
        related_name='files', on_delete=models.SET_NULL)
    scientist = models.ForeignKey(
        Person, blank=True, null=True, related_name='scientist',
        on_delete=models.SET_NULL, verbose_name=_(u"Scientist in charge"))
    research_comment = models.TextField(_(u"Research archaeology comment"),
                                        null=True, blank=True)
    classified_area = models.NullBooleanField(
        _(u"Classified area"), blank=True, null=True)
    protected_area = models.NullBooleanField(
        _(u"Protected area"), blank=True, null=True)
    if settings.COUNTRY == 'fr':
        cira_advised = models.NullBooleanField(
            u"Passage en CIRA", blank=True, null=True)
        mh_register = models.NullBooleanField(
            u"Sur Monument Historique classé", blank=True, null=True)
        mh_listing = models.NullBooleanField(
            u"Sur Monument Historique inscrit", blank=True, null=True)
    # <-- research archaeology
    cached_label = models.TextField(_(u"Cached name"), null=True, blank=True)
    imported_line = models.TextField(_(u"Imported line"), null=True,
                                     blank=True)
    history = HistoricalRecords()

    GET_VALUES_EXTRA = [
        'general_contractor_address_1',
        'general_contractor_address_2',
        'general_contractor_address_3',
        'get_locality',
    ]

    class Meta:
        verbose_name = _(u"Archaeological file")
        verbose_name_plural = _(u"Archaeological files")
        permissions = (
            ("view_file", ugettext(u"Can view all Archaeological files")),
            ("view_own_file", ugettext(u"Can view own Archaeological file")),
            ("add_own_file", ugettext(u"Can add own Archaeological file")),
            ("change_own_file",
             ugettext(u"Can change own Archaeological file")),
            ("delete_own_file",
             ugettext(u"Can delete own Archaeological file")),
            ("close_file", ugettext(u"Can close File")),
        )
        ordering = ('cached_label',)

    @property
    def short_class_name(self):
        return _(u"FILE")

    @property
    def full_internal_ref(self):
        return u"{}{}".format(settings.ISHTAR_FILE_PREFIX or '',
                              self.external_id or '')

    @property
    def delay_date(self):
        cache_key, val = get_cache(self.__class__, [self.pk, 'delay_date'])
        if val:
            return val
        return self.update_delay_date(cache_key)

    def update_delay_date(self, cache_key=None):
        if not cache_key:
            cache_key, val = get_cache(self.__class__,
                                       [self.pk, 'delay_date'])
        date = self.reception_date
        if not date:
            date = datetime.date(2500, 1, 1)
        elif settings.COUNTRY == 'fr' and self.saisine_type \
                and self.saisine_type.delay:
            date += datetime.timedelta(days=self.saisine_type.delay)
        cache.set(cache_key, date, settings.CACHE_TIMEOUT)
        return date

    @property
    def has_adminact(self):
        cache_key, val = get_cache(self.__class__, [self.pk,
                                                    'has_adminact'])
        if val:
            return val
        return self.update_has_admin_act(cache_key)

    @property
    def get_locality(self):
        return " - ".join(
            [getattr(self, k) for k in ('locality', 'address')
             if getattr(self, k)])

    @property
    def general_contractor_address_1(self):
        address = ''
        if self.general_contractor:
            if self.general_contractor.name:
                address = u" ".join([
                    unicode(getattr(self.general_contractor, key))
                    for key in ('title', 'surname', 'name')
                    if getattr(self.general_contractor, key)])
            elif self.general_contractor.raw_name:
                address = self.general_contractor.raw_name
        if not address and self.corporation_general_contractor and\
                self.corporation_general_contractor.name:
            address = self.corporation_general_contractor.name
        return address

    @property
    def general_contractor_address_2(self):
        address = ''
        if self.general_contractor and self.general_contractor.address:
            address = self.general_contractor.address
            if self.general_contractor.address_complement:
                address += u" " + self.general_contractor.address_complement
        if not address and self.corporation_general_contractor and\
                self.corporation_general_contractor.address:
            address = self.corporation_general_contractor.address
            if self.corporation_general_contractor.address_complement:
                address += u" " + \
                    self.corporation_general_contractor.address_complement
        return address

    @property
    def general_contractor_address_3(self):
        address = ''
        if self.general_contractor and self.general_contractor.postal_code:
            address = u" ".join([
                getattr(self.general_contractor, key)
                for key in ('postal_code', 'town')
                if getattr(self.general_contractor, key)])
        if not address and self.corporation_general_contractor and\
                self.corporation_general_contractor.address:
            address = u" ".join([
                getattr(self.corporation_general_contractor, key)
                for key in ('postal_code', 'town')
                if getattr(self.corporation_general_contractor, key)])
        return address

    @classmethod
    def similar_files(cls, parcels):
        # get similar parcels
        similar = set()
        for parcel in parcels:
            q = cls.objects.filter(
                parcels__town__pk=parcel['town'],
                parcels__section=parcel['section'],
                parcels__parcel_number=parcel['parcel_number']
            )
            if q.count():
                for fle in q.all():
                    similar.add(fle)
        return similar

    def update_has_admin_act(self, cache_key=None):
        if not cache_key:
            cache_key, val = get_cache(self.__class__, [self.pk,
                                                        'has_adminact'])
        has_adminact = self.administrative_act.exclude(
            act_type__txt_idx='a_receipt').count() or self.operations.count()
        cache.set(cache_key, has_adminact, settings.CACHE_TIMEOUT)
        return has_adminact

    @classmethod
    def get_short_menu_class(cls, pk):
        cache_key, val = get_cache(cls, [pk, 'short_class_name'])
        if val:
            return val
        q = cls.objects.filter(pk=pk)
        if not q.count():
            return ''
        item = q.all()[0]
        return item.update_short_menu_class(cache_key)

    def update_short_menu_class(self, cache_key=None):
        if not cache_key:
            cache_key, val = get_cache(self.__class__, [self.pk,
                                                        'short_class_name'])
        cls = 'normal'
        if not self.file_type.txt_idx == 'preventive':
            cls = "blue"
        elif not self.has_adminact and self.reception_date:
            delta = datetime.date.today() - self.reception_date
            cls = 'red'
            if self.saisine_type and self.saisine_type.delay:
                if delta.days <= (self.saisine_type.delay * 2 / 3):
                    cls = 'green'
                elif delta.days <= self.saisine_type.delay:
                    cls = 'orange'
        cache.set(cache_key, cls, settings.CACHE_TIMEOUT)
        return cls

    @classmethod
    def get_owns(cls, user, menu_filtr=None, limit=None, values=None,
                 get_short_menu_class=False):
        owns = super(File, cls).get_owns(
            user, limit=limit, values=values,
            get_short_menu_class=get_short_menu_class)
        return cls._return_get_owns(owns, values, get_short_menu_class)

    def get_values(self, prefix=''):
        values = super(File, self).get_values(prefix=prefix)
        return get_values_town_related(self, prefix, values)

    def render_parcels(self):
        from archaeological_operations.models import Parcel
        return Parcel.render_parcels(list(self.parcels.all()))

    def __unicode__(self):
        if self.cached_label:
            return self.cached_label
        self.save()
        return self.cached_label

    @property
    def short_label(self):
        return settings.JOINT.join(unicode(self).split(settings.JOINT)[1:])

    @property
    def reference(self):
        return self.external_id or u""

    def _generate_cached_label(self):
        items = [self.get_town_label(), self.reference]
        items += [unicode(getattr(self, k))
                  for k in ['internal_reference', 'name'] if getattr(self, k)]
        return settings.JOINT.join(items)

    def grouped_parcels(self):
        from archaeological_operations.models import Parcel
        return Parcel.grouped_parcels(list(self.parcels.all()))

    def get_town_label(self):
        lbl = unicode(_(u'Multi-town'))
        if self.main_town:
            lbl = self.main_town.name
        elif self.towns.count() == 1:
            lbl = self.towns.all()[0].name
        elif self.towns.count() == 0:
            lbl = unicode(_(u"No town"))
        return lbl

    def get_department(self):
        if not self.towns.count():
            return '00'
        return self.towns.all()[0].numero_insee[:2]

    @classmethod
    def get_query_owns(cls, user):
        return (Q(history_creator=user) |
                Q(in_charge__ishtaruser=user.ishtaruser)) \
            & Q(end_date__isnull=True)

    def is_active(self):
        return not bool(self.end_date)

    @property
    def town_list(self):
        return u", ".join([unicode(tw) for tw in self.towns.all()])

    def total_surface_ha(self):
        if self.total_surface:
            return self.total_surface / 10000.0

    def total_developed_surface_ha(self):
        if self.total_developed_surface:
            return self.total_developed_surface / 10000.0

    def operation_acts(self):
        acts = []
        for ope in self.operations.all():
            for act in ope.administrative_act.all():
                acts.append(act)
        return acts

    def update_raw_town_planning_service(self):
        if (self.raw_town_planning_service and not
            self.responsible_town_planning_service) or \
           not self.responsible_town_planning_service:
            return False
        current_lbl = ""
        if self.raw_town_planning_service:
            current_lbl = self.raw_town_planning_service[:]
        lbl = unicode(self.responsible_town_planning_service)
        if not lbl:
            return False
        self.raw_town_planning_service = lbl[:200]
        return current_lbl != self.raw_town_planning_service

    def update_planning_service(self):
        if not self.responsible_town_planning_service or \
           not self.responsible_town_planning_service.attached_to or \
           self.planning_service:
            return False
        self.planning_service = \
            self.responsible_town_planning_service.attached_to
        return True

    def update_resp_planning_service(self):
        if not self.responsible_town_planning_service or \
           self.responsible_town_planning_service.attached_to or \
           not self.planning_service:
            return False
        self.responsible_town_planning_service.attached_to = \
            self.planning_service
        self.responsible_town_planning_service.save()
        return True

    def update_raw_general_contractor(self):
        if (self.raw_general_contractor and not
           self.general_contractor) or \
           not self.general_contractor:
            return False
        current_lbl = ""
        if self.raw_general_contractor:
            current_lbl = self.raw_general_contractor[:]
        lbl = unicode(self.general_contractor)
        if not lbl:
            return False
        self.raw_general_contractor = lbl[:200]
        return current_lbl != self.raw_general_contractor

    def update_corpo_general_contractor(self):
        if not self.general_contractor or \
           self.general_contractor.attached_to \
           == self.corporation_general_contractor:
            return False
        if self.general_contractor.attached_to:
            self.corporation_general_contractor = \
                self.general_contractor.attached_to
        else:
            self.general_contractor.attached_to = \
                self.corporation_general_contractor
            self.general_contractor.save()
        return True

    def save(self, *args, **kwargs):
        returned = super(File, self).save(*args, **kwargs)
        if self.main_town and self.main_town not in list(self.towns.all()):
            self.towns.add(self.main_town)
        updated = self.update_raw_town_planning_service()
        updated += self.update_planning_service()
        self.update_resp_planning_service()
        updated += self.update_raw_general_contractor()
        updated += self.update_corpo_general_contractor()

        if not self.external_id or self.auto_external_id:
            external_id = get_external_id('file_external_id', self)
            if external_id != self.external_id:
                updated = True
                self.auto_external_id = True
                self.external_id = external_id
        if updated:
            self._cached_label_checked = False
            self.save()
            return returned

        self.update_delay_date()
        self.update_short_menu_class()
        return returned

    def is_preventive(self):
        return FileType.is_preventive(self.file_type.pk)

m2m_changed.connect(cached_label_changed, sender=File.towns.through)
post_save.connect(cached_label_changed, sender=File)


class FileByDepartment(models.Model):
    """
    Database view for dashboard
    """
    file = models.ForeignKey(File, verbose_name=_(u"File"))
    department = models.ForeignKey(Department, verbose_name=_(u"Department"),
                                   blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'file_department'

    def __unicode__(self):
        return u"{} - {}".format(self.file, self.department)


class FileDashboard:
    def __init__(self):
        from archaeological_operations.models import AdministrativeAct
        main_dashboard = Dashboard(File)

        self.total_number = main_dashboard.total_number

        types = File.objects.values('file_type', 'file_type__label')
        self.types = types.annotate(number=Count('pk')).order_by('file_type')

        by_year = File.objects.extra(
            {'date': "date_trunc('year', creation_date)"})
        self.by_year = by_year.values('date')\
                              .annotate(number=Count('pk')).order_by('-date')

        now = datetime.date.today()
        limit = datetime.date(now.year, now.month, 1) - datetime.timedelta(365)
        by_month = File.objects.filter(creation_date__gt=limit).extra(
            {'date': "date_trunc('month', creation_date)"})
        self.by_month = by_month.values('date')\
                                .annotate(number=Count('pk')).order_by('-date')

        # research
        self.research = {}
        prog_type = FileType.objects.get(txt_idx='prog')
        researchs = File.objects.filter(file_type=prog_type)
        self.research['total_number'] = researchs.count()
        by_year = researchs.extra(
            {'date': "date_trunc('year', creation_date)"}
        )
        self.research['by_year'] = by_year.values('date')\
                                          .annotate(number=Count('pk'))\
                                          .order_by('-date')
        by_month = researchs.filter(creation_date__gt=limit)\
            .extra({'date': "date_trunc('month', creation_date)"})
        self.research['by_month'] = by_month.values('date')\
                                            .annotate(number=Count('pk'))\
                                            .order_by('-date')

        self.research['by_dpt'] = FileByDepartment.objects\
            .filter(file__file_type=prog_type, department__isnull=False)\
            .values('department__label')\
            .annotate(number=Count('file'))\
            .order_by('department__label')
        FileTown = File.towns.through
        self.research['towns'] = FileTown.objects\
            .filter(file__file_type=prog_type)\
            .values('town__name')\
            .annotate(number=Count('file'))\
            .order_by('-number', 'town__name')[:10]

        # rescue
        rescue_type = FileType.objects.get(txt_idx='preventive')
        rescues = File.objects.filter(file_type=rescue_type)
        self.rescue = {}
        self.rescue['total_number'] = rescues.count()
        self.rescue['saisine'] = rescues.values('saisine_type__label')\
            .annotate(number=Count('pk')).order_by('saisine_type__label')
        self.rescue['administrative_act'] = AdministrativeAct.objects\
            .filter(associated_file__isnull=False)\
            .values('act_type__label')\
            .annotate(number=Count('pk'))\
            .order_by('act_type__pk')

        by_year = rescues.extra({'date': "date_trunc('year', creation_date)"})
        self.rescue['by_year'] = by_year.values('date')\
            .annotate(number=Count('pk')).order_by('-date')
        by_month = rescues.filter(creation_date__gt=limit)\
            .extra({'date': "date_trunc('month', creation_date)"})
        self.rescue['by_month'] = by_month.values('date')\
                                          .annotate(number=Count('pk'))\
                                          .order_by('-date')

        self.rescue['by_dpt'] = FileByDepartment.objects\
            .filter(file__file_type=rescue_type, department__isnull=False)\
            .values('department__label')\
            .annotate(number=Count('file'))\
            .order_by('department__label')
        self.rescue['towns'] = FileTown.objects\
            .filter(file__file_type=rescue_type)\
            .values('town__name')\
            .annotate(number=Count('file'))\
            .order_by('-number', 'town__name')[:10]

        self.rescue['with_associated_operation'] = rescues\
            .filter(operations__isnull=False).count()

        if self.rescue['total_number']:
            self.rescue['with_associated_operation_percent'] = round(
                float(self.rescue['with_associated_operation'])
                / self.rescue['total_number'] * 100, 2)

        by_year_operationnal = rescues.filter(operations__isnull=False)\
            .extra({'date': 'date_trunc(\'year\', '
                            '"archaeological_files_file".creation_date)'})
        by_year_operationnal = by_year_operationnal.values('date')\
            .annotate(number=Count('pk')).order_by('-date')
        percents, idx = [], 0
        for dct in self.rescue['by_year']:
            if idx >= len(by_year_operationnal):
                break
            if by_year_operationnal[idx]['date'] != dct['date'] or\
               not dct['number']:
                continue
            val = round(float(by_year_operationnal[idx]['number']) /
                        dct['number'] * 100, 2)
            percents.append({'date': dct['date'], 'number': val})
        self.rescue['operational_by_year'] = percents

        self.rescue['surface_by_town'] = FileTown.objects\
            .filter(file__file_type=rescue_type)\
            .values('town__name')\
            .annotate(number=Sum('file__total_surface'))\
            .order_by('-number', 'town__name')[:10]
        self.rescue['surface_by_dpt'] = FileByDepartment.objects\
            .filter(file__file_type=rescue_type, department__isnull=False)\
            .values('department__label')\
            .annotate(number=Sum('file__total_surface'))\
            .order_by('department__label')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2016  Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.utils.translation import ugettext_lazy as _

from ishtar_common.forms import reverse_lazy
from ishtar_common.wizards import Wizard, DeletionWizard, SourceWizard
from archaeological_operations.wizards import OperationAdministrativeActWizard

from archaeological_operations.models import AdministrativeAct
from archaeological_context_records.models import ContextRecord
import models


class FindWizard(Wizard):
    model = models.Find
    wizard_done_window = reverse_lazy('show-find')

    def get_current_contextrecord(self):
        step = self.steps.current
        if not step:
            return
        if step.endswith('_creation'):  # a context record has been selected
            main_form_key = 'selecrecord-' + self.url_name
            try:
                idx = int(self.session_get_value(main_form_key, 'pk'))
                current_cr = ContextRecord.objects.get(pk=idx)
                return current_cr
            except(TypeError, ValueError, ObjectDoesNotExist):
                pass
        current_item = self.get_current_object()
        if current_item:
            base_finds = current_item.base_finds.all()
            if base_finds:
                return base_finds[0].context_record

    def get_context_data(self, form, **kwargs):
        """
        Get the operation and context record "reminder" on top of wizard forms
        """
        context = super(FindWizard, self).get_context_data(form, **kwargs)
        current_cr = self.get_current_contextrecord()
        if not current_cr or self.steps.current.startswith('select-'):
            return context
        context['reminders'] = (
            (_(u"Operation"), unicode(current_cr.operation)),
            (_(u"Context record"), unicode(current_cr)))
        return context

    def get_extra_model(self, dct, form_list):
        dct = super(FindWizard, self).get_extra_model(dct, form_list)
        dct['order'] = 1
        if 'pk' in dct and type(dct['pk']) == ContextRecord:
            dct['base_finds__context_record'] = dct.pop('pk')
        return dct


class FindModificationWizard(FindWizard):
    modification = True
    main_item_select_keys = ('selec-', 'selecw-')
    filter_owns = {
        'selec-find_modification': ['pk'],
        'selecw-find_modification': ['pk'],
    }


class FindDeletionWizard(DeletionWizard):
    model = models.Find
    fields = ['label', 'material_types', 'datings', 'find_number',
              'object_types', 'description', 'conservatory_state', 'mark',
              'preservation_to_considers', 'integrities', 'remarkabilities',
              'volume', 'weight', 'length', 'width', 'height', 'diameter',
              'comment']


class TreatmentWizard(Wizard):
    model = models.Treatment
    wizard_done_window = reverse_lazy('show-treatment')
    basket_step = 'basetreatment-treatment_creation'
    saved_args = {"items": []}

    def get_form_kwargs(self, step, **kwargs):
        kwargs = super(TreatmentWizard, self).get_form_kwargs(step, **kwargs)
        if self.basket_step not in step:
            return kwargs
        kwargs['user'] = self.request.user
        return kwargs

    def get_extra_model(self, dct, form_list):
        """
        Get items concerned by the treatment
        """
        dct = super(TreatmentWizard, self).get_extra_model(dct, form_list)
        if 'resulting_pk' in dct:
            try:
                find = models.Find.objects.get(pk=dct.pop('resulting_pk'))
                if 'own' in self.current_right \
                        and not find.is_own(dct['history_modifier']):
                    raise PermissionDenied
                dct['items'] = [find]
            except models.Find.DoesNotExist:
                raise PermissionDenied
        if 'basket' in dct:
            basket = dct.pop('basket')
            if basket.user.pk != dct['history_modifier'].pk:
                raise PermissionDenied
            dct['items'] = list(basket.items.all())
        return dct


class TreatmentModificationWizard(TreatmentWizard):
    modification = True


class TreatmentDeletionWizard(DeletionWizard):
    model = models.Treatment
    fields = ['label', 'other_reference', 'year', 'index',
              'treatment_types', 'location', 'person', 'organization',
              'external_id', 'comment', 'description',
              'goal', 'start_date', 'end_date', 'container']


class TreatmentAdministrativeActWizard(OperationAdministrativeActWizard):
    model = models.Treatment
    current_obj_slug = 'administrativeacttreatment'
    ref_object_key = 'treatment'

    def get_reminder(self):
        return


class TreatmentEditAdministrativeActWizard(TreatmentAdministrativeActWizard):
    model = AdministrativeAct
    edit = True

    def get_associated_item(self, dct):
        return self.get_current_object().treatment


class TreatmentFileWizard(Wizard):
    model = models.TreatmentFile
    wizard_done_window = reverse_lazy('show-treatmentfile')


class TreatmentFileModificationWizard(TreatmentFileWizard):
    modification = True


class TreatmentFileDeletionWizard(DeletionWizard):
    model = models.TreatmentFile
    fields = ['name', 'internal_reference', 'external_id', 'year',
              'index', 'type', 'in_charge', 'reception_date',
              'creation_date', 'end_date', 'comment']


class TreatmentFileAdministrativeActWizard(
        OperationAdministrativeActWizard):
    model = models.TreatmentFile
    current_obj_slug = 'administrativeacttreatmentfile'
    ref_object_key = 'treatment_file'

    def get_reminder(self):
        form_key = 'selec-' + self.url_name
        if self.url_name.endswith('_administrativeactop'):
            # modification and deletion are suffixed with '_modification'
            # and '_deletion' so it is creation
            pk = self.session_get_value(form_key, "pk")
            try:
                return (
                    (_(u"Treatment request"),
                     unicode(models.TreatmentFile.objects.get(pk=pk))),
                )
            except models.TreatmentFile.DoesNotExist:
                return
        else:
            admin_id = self.session_get_value(form_key, "pk")
            try:
                admin = AdministrativeAct.objects.get(pk=admin_id)
                if not admin.operation:
                    return
                return (
                    (_(u"Operation"), unicode(admin.operation)),
                )
            except AdministrativeAct.DoesNotExist:
                return
        return


class TreatmentFileEditAdministrativeActWizard(
        TreatmentFileAdministrativeActWizard):
    model = AdministrativeAct
    edit = True

    def get_associated_item(self, dct):
        return self.get_current_object().treatment_file


class FindSourceWizard(SourceWizard):
    wizard_done_window = reverse_lazy('show-findsource')
    model = models.FindSource


class FindSourceDeletionWizard(DeletionWizard):
    model = models.FindSource
    fields = ['item', 'title', 'source_type', 'authors', ]


class TreatmentSourceWizard(SourceWizard):
    model = models.TreatmentSource


class TreatmentSourceDeletionWizard(DeletionWizard):
    model = models.TreatmentSource
    fields = ['treatment', 'title', 'source_type', 'authors']


class TreatmentFileSourceWizard(SourceWizard):
    model = models.TreatmentFileSource


class TreatmentFileSourceDeletionWizard(DeletionWizard):
    model = models.TreatmentFileSource
    fields = ['treatment_file', 'title', 'source_type', 'authors']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2016  Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

import datetime
import logging

from django import forms
from django.conf import settings
from django.core import validators
from django.db.models import Max
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from ishtar_common.models import Person, valid_id, Organization, \
    get_current_profile
from archaeological_operations.models import ActType, AdministrativeAct
from archaeological_warehouse.models import Warehouse, Container
import models

from archaeological_operations.forms import AdministrativeActOpeForm, \
    AdministrativeActOpeFormSelection, AdministrativeActModifForm

from ishtar_common.forms import reverse_lazy, TableSelect, FinalForm, \
    ManageOldType, get_form_selection
from ishtar_common.forms_common import SourceForm, SourceSelect, \
    SourceDeletionForm

from ishtar_common import widgets

logger = logging.getLogger(__name__)

# Treatment


class TreatmentSelect(TableSelect):
    label = forms.CharField(label=_(u"Label"))
    other_reference = forms.CharField(label=_(u"Other ref."))
    year = forms.IntegerField(label=_(u"Year"))
    index = forms.IntegerField(label=_(u"Index"))
    treatment_types = forms.ChoiceField(label=_(u"Treatment type"), choices=[])
    image = forms.NullBooleanField(label=_(u"Has an image?"))

    def __init__(self, *args, **kwargs):
        super(TreatmentSelect, self).__init__(*args, **kwargs)
        self.fields['treatment_types'].choices = \
            models.TreatmentType.get_types()
        self.fields['treatment_types'].help_text = \
            models.TreatmentType.get_help()


class TreatmentFormSelection(forms.Form):
    form_label = _("Treatment search")
    associated_models = {'pk': models.Treatment}
    currents = {'pk': models.Treatment}
    pk = forms.IntegerField(
        label="", required=False,
        widget=widgets.JQueryJqGrid(
            reverse_lazy('get-treatment'),
            TreatmentSelect, models.Treatment),
        validators=[valid_id(models.Treatment)])


class BaseTreatmentForm(ManageOldType, forms.Form):
    form_label = _(u"Base treatment")
    base_models = ['treatment_type']
    associated_models = {'treatment_type': models.TreatmentType,
                         'person': Person,
                         'location': Warehouse,
                         'organization': Organization,
                         'container': Container,
                         'treatment_state': models.TreatmentState,
                         }
    file_upload = True
    need_user_for_initialization = True

    label = forms.CharField(label=_(u"Label"),
                            max_length=200, required=False)
    other_reference = forms.CharField(
        label=_(u"Other ref."), max_length=200, required=False)
    year = forms.IntegerField(label=_("Year"),
                              initial=lambda: datetime.datetime.now().year,
                              validators=[validators.MinValueValidator(1600),
                                          validators.MaxValueValidator(2100)])
    treatment_type = forms.MultipleChoiceField(
        label=_(u"Treatment type"), choices=[],
        widget=widgets.CheckboxSelectMultiple)
    treatment_state = forms.ChoiceField(label=_(u"State"),
                                        choices=[], required=False)
    target_is_basket = forms.NullBooleanField(label=_(u"Target"))
    person = forms.IntegerField(
        label=_(u"Responsible"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person'), associated_model=Person,
            new=True),
        validators=[valid_id(Person)], required=False)
    organization = forms.IntegerField(
        label=_(u"Organization"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-organization'),
            associated_model=Organization, new=True),
        validators=[valid_id(Organization)], required=False)
    location = forms.IntegerField(
        label=_(u"Location"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-warehouse'), associated_model=Warehouse,
            new=True),
        validators=[valid_id(Warehouse)])
    container = forms.IntegerField(
        label=_(u"Container (relevant for packaging)"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-container'),
            associated_model=Container, new=True),
        validators=[valid_id(Container)], required=False)
    external_id = forms.CharField(
        label=_(u"External ref."), max_length=200, required=False)
    comment = forms.CharField(label=_(u"Comment"),
                              widget=forms.Textarea, required=False)
    description = forms.CharField(label=_(u"Description"),
                                  widget=forms.Textarea, required=False)
    goal = forms.CharField(label=_(u"Goal"),
                           widget=forms.Textarea, required=False)
    start_date = forms.DateField(label=_(u"Start date"), required=False,
                                 widget=widgets.JQueryDate)
    end_date = forms.DateField(label=_(u"Closing date"), required=False,
                               widget=widgets.JQueryDate)
    estimated_cost = forms.FloatField(label=_(u"Estimated cost ({currency})"),
                                      required=False)
    quoted_cost = forms.FloatField(label=_(u"Quoted cost ({currency})"),
                                   required=False)
    realized_cost = forms.FloatField(label=_(u"Realized cost ({currency})"),
                                     required=False)
    insurance_cost = forms.FloatField(label=_(u"Insurance cost ({currency})"),
                                      required=False)
    image = forms.ImageField(
        label=_(u"Image"), help_text=mark_safe(
            _(u"<p>Heavy images are resized to: %(width)dx%(height)d "
              u"(ratio is preserved).</p>") % {
                'width': settings.IMAGE_MAX_SIZE[0],
                'height': settings.IMAGE_MAX_SIZE[1]}),
        max_length=255, required=False, widget=widgets.ImageFileInput())

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user')
        super(BaseTreatmentForm, self).__init__(*args, **kwargs)
        # set current currency
        currency = get_current_profile().currency
        for key in ('estimated_cost', 'quoted_cost', 'realized_cost',
                    'insurance_cost'):
            self.fields[key].label = self.fields[key].label.format(
                currency=currency)

        q = Person.objects.filter(ishtaruser__pk=user.pk)
        if q.count():
            person = q.all()[0]
            self.fields['person'].initial = person.pk
            if person.attached_to:
                self.fields['organization'].initial = person.attached_to.pk
        self.fields['target_is_basket'].widget.choices = \
            ((False, _(u"Single find")), (True, _(u"Basket")))
        self.fields['treatment_type'].choices = models.TreatmentType.get_types(
            initial=self.init_data.get('treatment_type'),
            dct={'upstream_is_many': False, 'downstream_is_many': False},
            empty_first=False
        )
        self.fields['treatment_type'].help_text = \
            models.TreatmentType.get_help(
                dct={'upstream_is_many': False, 'downstream_is_many': False})
        self.fields['treatment_state'].choices = \
            models.TreatmentState.get_types(
                initial=self.init_data.get('treatment_state'),
            )
        self.fields['treatment_state'].help_text = \
            models.TreatmentState.get_help()
        # TODO
        """
        self.fields['basket'].required = False
        self.fields['basket'].help_text = \
            _(u"Leave it blank if you want to select a single item")
        self.fields.keyOrder.pop(self.fields.keyOrder.index('basket'))
        self.fields.keyOrder.insert(self.fields.keyOrder.index('description'),
                                    'basket')
        """

    def clean(self, *args, **kwargs):
        data = self.cleaned_data
        packaging = models.TreatmentType.get_cache('packaging')
        if not packaging:
            logger.warning("No 'packaging' treatment type defined")
            return
        if data.get('container', None) \
                and str(packaging.pk) not in data.get('treatment_type', []):
            raise forms.ValidationError(
                _(u"The container field is attached to the treatment. If "
                  u"no packaging treatment is done it is not relevant."))
        if not data.get('container', None) \
                and str(packaging.pk) in data.get('treatment_type', []):
            raise forms.ValidationError(
                _(u"If a packaging treatment is done, the container field "
                  u"must be filled."))
        if not data.get('person', None) and not data.get('organization', None):
            raise forms.ValidationError(
                _(u"A responsible or an organization must be defined."))
        return data
        # TODO
        """
        for treatment_type in self.cleaned_data.get('treatment_type', []):
            try:
                treatment = models.TreatmentType.objects.get(
                    pk=treatment_type, available=True)
            except models.TreatmentType.DoesNotExist:
                raise forms.ValidationError(_(u"This treatment type is not "
                                              u"available."))
            if treatment.upstream_is_many and \
                    not self.cleaned_data.get('basket'):
                raise forms.ValidationError(_(u"This treatment needs a "
                                              u"basket."))
        """


class TreatmentModifyForm(BaseTreatmentForm):
    index = forms.IntegerField(_(u"Index"))
    id = forms.IntegerField(' ', widget=forms.HiddenInput, required=False)

    def __init__(self, *args, **kwargs):
        super(TreatmentModifyForm, self).__init__(*args, **kwargs)
        self.fields.keyOrder.pop(self.fields.keyOrder.index('index'))
        self.fields.keyOrder.insert(
            self.fields.keyOrder.index('year') + 1, 'index')

    def clean(self, *args, **kwargs):
        super(TreatmentModifyForm, self).clean(*args, **kwargs)
        cleaned_data = self.cleaned_data
        year = cleaned_data.get('year')
        pk = cleaned_data.get('id')
        index = cleaned_data.get('index')
        q = models.Treatment.objects \
            .filter(year=year, index=index).exclude(pk=pk)
        if index and q.count():
            raise forms.ValidationError(
                _(u"Another treatment with this index exists for {}."
                  ).format(year))
        return cleaned_data


class TreatmentFormFileChoice(forms.Form):
    form_label = _(u"Associated request")
    associated_models = {'file': models.TreatmentFile, }
    currents = {'file': models.TreatmentFile}
    file = forms.IntegerField(
        label=_(u"Treatment request"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-treatmentfile'),
            associated_model=models.TreatmentFile),
        validators=[valid_id(models.TreatmentFile)], required=False)


class TreatmentDeletionForm(FinalForm):
    confirm_msg = _(
        u"Are you sure you want to delete this treatment? All changes "
        u"made to the associated finds since this treatment record will be "
        u"lost!")
    confirm_end_msg = _(u"Would you like to delete this treatment?")

# administrative act treatment


class AdministrativeActTreatmentSelect(TableSelect):
    year = forms.IntegerField(label=_("Year"))
    index = forms.IntegerField(label=_("Index"))
    act_type = forms.ChoiceField(label=_("Act type"), choices=[])
    indexed = forms.NullBooleanField(label=_(u"Indexed?"))
    act_object = forms.CharField(label=_(u"Object"),
                                 max_length=300)

    signature_date_after = forms.DateField(
        label=_(u"Signature date after"), widget=widgets.JQueryDate)
    signature_date_before = forms.DateField(
        label=_(u"Signature date before"), widget=widgets.JQueryDate)
    treatment__name = forms.CharField(
        label=_(u"Treatment name"), max_length=200)
    treatment__year = forms.IntegerField(label=_(u"Treatment year"))
    treatment__index = forms.IntegerField(label=_(u"Treatment index"))
    treatment__internal_reference = forms.CharField(
        max_length=200, label=_(u"Treatment internal reference"))
    treatment__treatment_types = forms.ChoiceField(label=_(u"Treatment type"),
                                                   choices=[])
    history_modifier = forms.IntegerField(
        label=_(u"Modified by"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person',
                         args=['0', 'user']),
            associated_model=Person),
        validators=[valid_id(Person)])

    def __init__(self, *args, **kwargs):
        super(AdministrativeActTreatmentSelect, self).__init__(*args, **kwargs)
        self.fields['act_type'].choices = ActType.get_types(
            dct={'intented_to': 'T'})
        self.fields['act_type'].help_text = ActType.get_help(
            dct={'intented_to': 'T'})
        self.fields['treatment__treatment_types'].choices = \
            models.TreatmentType.get_types()
        self.fields['treatment__treatment_types'].help_text = \
            models.TreatmentType.get_help()


class AdministrativeActTreatmentFormSelection(
        AdministrativeActOpeFormSelection):
    pk = forms.IntegerField(
        label="", required=False,
        widget=widgets.JQueryJqGrid(
            reverse_lazy('get-administrativeacttreatment'),
            AdministrativeActTreatmentSelect, AdministrativeAct),
        validators=[valid_id(AdministrativeAct)])


class AdministrativeActTreatmentForm(AdministrativeActOpeForm):
    act_type = forms.ChoiceField(label=_(u"Act type"), choices=[])

    def __init__(self, *args, **kwargs):
        super(AdministrativeActTreatmentForm, self).__init__(*args, **kwargs)
        self.fields['act_type'].choices = ActType.get_types(
            initial=self.init_data.get('act_type'),
            dct={'intented_to': 'T'})
        self.fields['act_type'].help_text = ActType.get_help(
            dct={'intented_to': 'T'})


class AdministrativeActTreatmentModifForm(
        AdministrativeActModifForm, AdministrativeActTreatmentForm):
    pk = forms.IntegerField(required=False, widget=forms.HiddenInput)
    index = forms.IntegerField(label=_("Index"), required=False)

# treatment requests


class TreatmentFileSelect(TableSelect):
    name = forms.CharField(label=_(u"Name"))
    internal_reference = forms.CharField(label=_(u"Internal ref."))
    year = forms.IntegerField(label=_(u"Year"))
    index = forms.IntegerField(label=_(u"Index"))
    type = forms.ChoiceField(label=_(u"Type"), choices=[])
    in_charge = forms.IntegerField(
        label=_(u"In charge"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person'),
            associated_model=Person),
        validators=[valid_id(Person)])
    applicant = forms.IntegerField(
        label=_(u"Applicant"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person'),
            associated_model=Person),
        validators=[valid_id(Person)])
    applicant_organisation = forms.IntegerField(
        label=_(u"Applicant organisation"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-organization'),
            associated_model=Organization),
        validators=[valid_id(Organization)])

    def __init__(self, *args, **kwargs):
        super(TreatmentFileSelect, self).__init__(*args, **kwargs)
        self.fields['type'].choices = models.TreatmentFileType.get_types()
        self.fields['type'].help_text = models.TreatmentFileType.get_help()


class TreatmentFileFormSelection(forms.Form):
    form_label = _("Treatment request search")
    associated_models = {'pk': models.TreatmentFile}
    currents = {'pk': models.TreatmentFile}
    pk = forms.IntegerField(
        label="", required=False,
        widget=widgets.JQueryJqGrid(
            reverse_lazy('get-treatmentfile'),
            TreatmentFileSelect, models.TreatmentFile),
        validators=[valid_id(models.TreatmentFile)])


class TreatmentFileForm(ManageOldType, forms.Form):
    form_label = _(u"Treatment request")
    base_models = ['treatment_type_type']
    associated_models = {
        'type': models.TreatmentFileType, 'in_charge': Person,
        'applicant': Person, 'applicant_organisation': Organization}
    need_user_for_initialization = True

    name = forms.CharField(label=_(u"Name"),
                           max_length=1000, required=False)
    year = forms.IntegerField(label=_("Year"),
                              initial=lambda: datetime.datetime.now().year,
                              validators=[validators.MinValueValidator(1600),
                                          validators.MaxValueValidator(2100)])
    internal_reference = forms.CharField(
        label=_(u"Internal ref."), max_length=60, required=False)
    external_id = forms.CharField(
        label=_(u"External ref."), max_length=200, required=False)
    type = forms.ChoiceField(
        label=_(u"Type"), choices=[])
    in_charge = forms.IntegerField(
        label=_(u"Responsible"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person'), associated_model=Person,
            new=True),
        validators=[valid_id(Person)])
    applicant = forms.IntegerField(
        label=_(u"Applicant"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person'), associated_model=Person,
            new=True),
        validators=[valid_id(Person)], required=False)
    applicant_organisation = forms.IntegerField(
        label=_(u"Applicant organisation"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-organization'),
            associated_model=Organization, new=True),
        validators=[valid_id(Organization)], required=False)
    comment = forms.CharField(label=_(u"Comment"),
                              widget=forms.Textarea, required=False)
    creation_date = forms.DateField(label=_(u"Start date"), required=False,
                                    widget=widgets.JQueryDate,
                                    initial=lambda: datetime.datetime.now())
    reception_date = forms.DateField(label=_(u"Reception date"), required=False,
                                     widget=widgets.JQueryDate,
                                     initial=lambda: datetime.datetime.now())
    end_date = forms.DateField(label=_(u"Closing date"), required=False,
                               widget=widgets.JQueryDate)

    def __init__(self, *args, **kwargs):
        user = kwargs.pop('user')
        super(TreatmentFileForm, self).__init__(*args, **kwargs)
        q = Person.objects.filter(ishtaruser__pk=user.pk)
        if q.count():
            person = q.all()[0]
            self.fields['in_charge'].initial = person.pk
        self.fields['type'].choices = models.TreatmentFileType.get_types(
            initial=[self.init_data.get('type')], empty_first=False
        )
        self.fields['type'].help_text = models.TreatmentFileType.get_help()


class TreatmentFileModifyForm(TreatmentFileForm):
    index = forms.IntegerField(_(u"Index"))
    id = forms.IntegerField(' ', widget=forms.HiddenInput, required=False)

    def __init__(self, *args, **kwargs):
        super(TreatmentFileModifyForm, self).__init__(*args, **kwargs)
        self.fields.keyOrder.pop(self.fields.keyOrder.index('index'))
        self.fields.keyOrder.insert(
            self.fields.keyOrder.index('year') + 1, 'index')

    def clean(self, *args, **kwargs):
        super(TreatmentFileModifyForm, self).clean(*args, **kwargs)
        cleaned_data = self.cleaned_data
        year = cleaned_data.get('year')
        pk = cleaned_data.get('id')
        index = cleaned_data.get('index')
        q = models.TreatmentFile.objects \
            .filter(year=year, index=index).exclude(pk=pk)
        if index and q.count():
            raise forms.ValidationError(
                _(u"Another treatment request with this index exists for {}."
                  ).format(year))
        return cleaned_data


class TreatmentFileDeletionForm(FinalForm):
    confirm_msg = _(u"Are you sure you want to delete this treatment request?")
    confirm_end_msg = _(u"Would you like to delete this treatment request?")


class AdministrativeActTreatmentFileSelect(TableSelect):
    year = forms.IntegerField(label=_("Year"))
    index = forms.IntegerField(label=_("Index"))
    act_type = forms.ChoiceField(label=_("Act type"), choices=[])
    indexed = forms.NullBooleanField(label=_(u"Indexed?"))
    act_object = forms.CharField(label=_(u"Object"),
                                 max_length=300)

    signature_date_after = forms.DateField(
        label=_(u"Signature date after"), widget=widgets.JQueryDate)
    signature_date_before = forms.DateField(
        label=_(u"Signature date before"), widget=widgets.JQueryDate)
    treatment_file__name = forms.CharField(
        label=_(u"Treatment request name"), max_length=200)
    treatment_file__year = forms.IntegerField(
        label=_(u"Treatment request year"))
    treatment_file__index = forms.IntegerField(
        label=_(u"Treatment request index"))
    treatment_file__internal_reference = forms.CharField(
        max_length=200, label=_(u"Treatment request internal reference"))
    treatment_file__type = forms.ChoiceField(label=_(u"Treatment request type"),
                                             choices=[])
    history_modifier = forms.IntegerField(
        label=_(u"Modified by"),
        widget=widgets.JQueryAutoComplete(
            reverse_lazy('autocomplete-person',
                         args=['0', 'user']),
            associated_model=Person),
        validators=[valid_id(Person)])

    def __init__(self, *args, **kwargs):
        super(AdministrativeActTreatmentFileSelect, self).__init__(*args,
                                                                   **kwargs)
        self.fields['act_type'].choices = ActType.get_types(
            dct={'intented_to': 'TF'})
        self.fields['act_type'].help_text = ActType.get_help(
            dct={'intented_to': 'TF'})
        self.fields['treatment_file__type'].choices = \
            models.TreatmentFileType.get_types()
        self.fields['treatment_file__type'].help_text = \
            models.TreatmentFileType.get_help()


class AdministrativeActTreatmentFileFormSelection(
        AdministrativeActOpeFormSelection):
    pk = forms.IntegerField(
        label="", required=False,
        widget=widgets.JQueryJqGrid(
            reverse_lazy('get-administrativeacttreatmentfile'),
            AdministrativeActTreatmentFileSelect, AdministrativeAct),
        validators=[valid_id(AdministrativeAct)])


class AdministrativeActTreatmentFileForm(AdministrativeActOpeForm):
    act_type = forms.ChoiceField(label=_(u"Act type"), choices=[])

    def __init__(self, *args, **kwargs):
        super(AdministrativeActTreatmentFileForm, self).__init__(*args,
                                                                 **kwargs)
        self.fields['act_type'].choices = ActType.get_types(
            initial=self.init_data.get('act_type'),
            dct={'intented_to': 'TF'})
        self.fields['act_type'].help_text = ActType.get_help(
            dct={'intented_to': 'TF'})


class AdministrativeActTreatmentFileModifForm(
        AdministrativeActModifForm, AdministrativeActTreatmentFileForm):
    pk = forms.IntegerField(required=False, widget=forms.HiddenInput)
    index = forms.IntegerField(label=_("Index"), required=False)

SourceTreatmentFormSelection = get_form_selection(
    'SourceTreatmentFormSelection', _(u"Treatment search"), 'treatment',
    models.Treatment, TreatmentSelect, 'get-treatment',
    _(u"You should select a treatment."))

SourceTreatmentFileFormSelection = get_form_selection(
    'SourceTreatmentFileFormSelection', _(u"Treatment request search"),
    'treatment_file',
    models.TreatmentFile, TreatmentFileSelect, 'get-treatmentfile',
    _(u"You should select a treatment request."))


class TreatmentSourceSelect(SourceSelect):
    treatment__name = forms.CharField(
        label=_(u"Treatment name"), max_length=200)
    treatment__year = forms.IntegerField(label=_(u"Treatment year"))
    treatment__index = forms.IntegerField(label=_(u"Treatment index"))
    treatment__internal_reference = forms.CharField(
        max_length=200, label=_(u"Treatment internal reference"))
    treatment__treatment_types = forms.ChoiceField(label=_(u"Treatment type"),
                                                   choices=[])

    def __init__(self, *args, **kwargs):
        super(TreatmentSourceSelect, self).__init__(*args, **kwargs)
        self.fields['treatment__treatment_types'].choices = \
            models.TreatmentType.get_types()
        self.fields['treatment__treatment_types'].help_text = \
            models.TreatmentType.get_help()


TreatmentSourceFormSelection = get_form_selection(
    'TreatmentSourceFormSelection', _(u"Documentation search"), 'pk',
    models.TreatmentSource, TreatmentSourceSelect, 'get-treatmentsource',
    _(u"You should select a document."))


class TreatmentFileSourceSelect(SourceSelect):
    treatment_file__name = forms.CharField(
        label=_(u"Treatment request name"), max_length=200)
    treatment_file__year = forms.IntegerField(
        label=_(u"Treatment request year"))
    treatment_file__index = forms.IntegerField(
        label=_(u"Treatment request index"))
    treatment_file__internal_reference = forms.CharField(
        max_length=200, label=_(u"Treatment request internal reference"))
    treatment_file__type = forms.ChoiceField(label=_(u"Treatment request type"),
                                             choices=[])

    def __init__(self, *args, **kwargs):
        super(TreatmentFileSourceSelect, self).__init__(*args, **kwargs)
        self.fields['treatment_file__type'].choices = \
            models.TreatmentFileType.get_types()
        self.fields['treatment_file__type'].help_text = \
            models.TreatmentFileType.get_help()


TreatmentFileSourceFormSelection = get_form_selection(
    'TreatmentFileSourceFormSelection', _(u"Documentation search"), 'pk',
    models.TreatmentFileSource, TreatmentFileSourceSelect,
    'get-treatmentfilesource', _(u"You should select a document."))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2017 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

import datetime

from django.conf import settings
from django.contrib.gis.db import models
from django.db.models import Q, Count
from django.db.models.signals import post_save, post_delete
from django.template.defaultfilters import slugify
from django.utils.translation import ugettext_lazy as _, ugettext

from ishtar_common.utils import cached_label_changed

from ishtar_common.models import GeneralType, get_external_id, \
    LightHistorizedItem, OwnPerms, Address, Person, post_save_cache, \
    ImageModel, DashboardFormItem


class WarehouseType(GeneralType):
    class Meta:
        verbose_name = _(u"Warehouse type")
        verbose_name_plural = _(u"Warehouse types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=WarehouseType)
post_delete.connect(post_save_cache, sender=WarehouseType)


class Warehouse(Address, DashboardFormItem, OwnPerms):
    SHOW_URL = 'show-warehouse'
    name = models.CharField(_(u"Name"), max_length=200)
    warehouse_type = models.ForeignKey(WarehouseType,
                                       verbose_name=_(u"Warehouse type"))
    person_in_charge = models.ForeignKey(
        Person, on_delete=models.SET_NULL, related_name='warehouse_in_charge',
        verbose_name=_(u"Person in charge"), null=True, blank=True)
    comment = models.TextField(_(u"Comment"), null=True, blank=True)
    associated_divisions = models.ManyToManyField(
        'WarehouseDivision', verbose_name=_("Divisions"), blank=True,
        through='WarehouseDivisionLink'
    )
    external_id = models.TextField(_(u"External ID"), blank=True, null=True)
    auto_external_id = models.BooleanField(
        _(u"External ID is set automatically"), default=False)

    TABLE_COLS = ['name', 'warehouse_type']

    class Meta:
        verbose_name = _(u"Warehouse")
        verbose_name_plural = _(u"Warehouses")
        permissions = (
            ("view_warehouse", ugettext(u"Can view all Warehouses")),
            ("view_own_warehouse", ugettext(u"Can view own Warehouse")),
            ("add_own_warehouse", ugettext(u"Can add own Warehouse")),
            ("change_own_warehouse", ugettext(u"Can change own Warehouse")),
            ("delete_own_warehouse", ugettext(u"Can delete own Warehouse")),
        )

    def __unicode__(self):
        return u"%s (%s)" % (self.name, unicode(self.warehouse_type))

    @property
    def associated_filename(self):
        return datetime.date.today().strftime('%Y-%m-%d') + '-' + \
               slugify(unicode(self))

    @classmethod
    def get_query_owns(cls, user):
        return Q(person_in_charge__ishtaruser=user.ishtaruser)

    @property
    def number_of_finds(self):
        from archaeological_finds.models import Find
        return Find.objects.filter(container__responsible=self).count()

    @property
    def number_of_finds_hosted(self):
        from archaeological_finds.models import Find
        return Find.objects.filter(container__location=self).count()

    @property
    def number_of_containers(self):
        return Container.objects.filter(location=self).count()

    def _get_divisions(self, current_path, remaining_division, depth=0):
        if not remaining_division:
            return [current_path]
        current_division = remaining_division.pop(0)
        q = ContainerLocalisation.objects.filter(
            division=current_division,
        )
        for div, ref in current_path:
            q = q.filter(
                container__division__division=div,
                container__division__reference=ref
            )
        res = []
        old_ref = None
        for ref in q.values('reference').order_by('reference').all():
            if ref['reference'] == old_ref:
                continue
            old_ref = ref['reference']
            cpath = current_path[:]
            cpath.append((current_division, ref['reference']))
            for r in self._get_divisions(cpath, remaining_division[:],
                                         depth + 1):
                res.append(r)
        return res

    @property
    def available_division_tuples(self):
        """
        :return: ordered list of available paths. Each path is a list of
        tuple with the WarehouseDivisionLink and the reference.
        """
        divisions = list(
            WarehouseDivisionLink.objects.filter(warehouse=self
                                                 ).order_by('order').all())
        return self._get_divisions([], divisions)

    def _number_of_items_by_place(self, model, division_key='division'):
        res = {}
        paths = self.available_division_tuples[:]
        for path in paths:
            q = model.objects
            cpath = []
            for division, ref in path:
                lbl = u"{} {}".format(division.division, ref)
                cpath.append(lbl)
                attrs = {
                    division_key + "__division": division,
                    division_key + "__reference": ref
                }
                q = q.filter(**attrs)
                if tuple(cpath) not in res:
                    res[tuple(cpath)] = q.count()
        res = [(k, res[k]) for k in res]
        final_res, current_res, depth = [], [], 1
        for path, nb in sorted(res, key=lambda x: (len(x[0]), x[0])):
            if depth != len(path):
                final_res.append(current_res[:])
                current_res = []
                depth = len(path)
            current_res.append((u" | ".join(path), nb))
        final_res.append(current_res[:])
        return final_res

    def _number_of_finds_by_place(self):
        from archaeological_finds.models import Find
        return self._number_of_items_by_place(
            Find, division_key='container__division')

    @property
    def number_of_finds_by_place(self, update=False):
        return self._get_or_set_stats('_number_of_finds_by_place', update,
                                      settings.CACHE_SMALLTIMEOUT)

    def _number_of_containers_by_place(self):
        return self._number_of_items_by_place(Container)

    @property
    def number_of_containers_by_place(self, update=False):
        return self._get_or_set_stats('_number_of_containers_by_place', update,
                                      settings.CACHE_SMALLTIMEOUT)

    def save(self, *args, **kwargs):
        super(Warehouse, self).save(*args, **kwargs)
        for container in self.containers.all():
            cached_label_changed(Container, instance=container)

        self.skip_history_when_saving = True
        if not self.external_id or self.auto_external_id:
            external_id = get_external_id('warehouse_external_id', self)
            if external_id != self.external_id:
                updated = True
                self.auto_external_id = True
                self.external_id = external_id

                self._cached_label_checked = False
                self.save()
                return


class Collection(LightHistorizedItem):
    name = models.CharField(_(u"Name"), max_length=200,
                            null=True, blank=True)
    description = models.TextField(_(u"Description"), null=True, blank=True)
    warehouse = models.ForeignKey(Warehouse, verbose_name=_(u"Warehouse"),
                                  related_name='collections')

    class Meta:
        verbose_name = _(u"Collection")
        verbose_name_plural = _(u"Collection")
        ordering = ('name',)

    def __unicode__(self):
        return self.name


class WarehouseDivision(GeneralType):
    class Meta:
        verbose_name = _(u"Warehouse division type")
        verbose_name_plural = _(u"Warehouse division types")
post_save.connect(post_save_cache, sender=WarehouseDivision)
post_delete.connect(post_save_cache, sender=WarehouseDivision)


class WarehouseDivisionLink(models.Model):
    RELATIVE_MODELS = {Warehouse: 'warehouse'}
    warehouse = models.ForeignKey(Warehouse)
    division = models.ForeignKey(WarehouseDivision)
    order = models.IntegerField(_("Order"), default=10)

    class Meta:
        ordering = ('warehouse', 'order')
        unique_together = ('warehouse', 'division')

    def __unicode__(self):
        return u"{} - {}".format(self.warehouse, self.division)


class ContainerType(GeneralType):
    length = models.IntegerField(_(u"Length (mm)"), blank=True, null=True)
    width = models.IntegerField(_(u"Width (mm)"), blank=True, null=True)
    height = models.IntegerField(_(u"Height (mm)"), blank=True, null=True)
    volume = models.FloatField(_(u"Volume (l)"), blank=True, null=True)
    reference = models.CharField(_(u"Ref."), max_length=30)

    class Meta:
        verbose_name = _(u"Container type")
        verbose_name_plural = _(u"Container types")
        ordering = ('label',)
post_save.connect(post_save_cache, sender=ContainerType)
post_delete.connect(post_save_cache, sender=ContainerType)


class Container(LightHistorizedItem, ImageModel):
    TABLE_COLS = ['reference', 'container_type__label', 'cached_location',
                  'divisions_lbl']
    IMAGE_PREFIX = 'containers/'

    # search parameters
    EXTRA_REQUEST_KEYS = {
        'location': 'location__pk',
        'container_type': 'container_type__pk',
        'reference': 'reference__icontains',
    }
    SHOW_URL = 'show-container'
    COL_LABELS = {
        'cached_location': _(u"Location - index"),
        'divisions_lbl': _(u"Precise localisation"),
        'container_type__label': _(u"Type")
    }
    CACHED_LABELS = ['cached_label', 'cached_location']

    # fields
    location = models.ForeignKey(
        Warehouse, verbose_name=_(u"Location (warehouse)"),
        related_name='containers')
    responsible = models.ForeignKey(
        Warehouse, verbose_name=_(u"Responsible warehouse"),
        related_name='owned_containers')
    container_type = models.ForeignKey(ContainerType,
                                       verbose_name=_("Container type"))
    reference = models.CharField(_(u"Container ref."), max_length=40)
    comment = models.TextField(_(u"Comment"), null=True, blank=True)
    cached_label = models.CharField(_(u"Localisation"), max_length=500,
                                    null=True, blank=True)
    cached_location = models.CharField(_(u"Cached location"), max_length=500,
                                       null=True, blank=True)
    index = models.IntegerField(u"ID", default=0)
    external_id = models.TextField(_(u"External ID"), blank=True, null=True)
    auto_external_id = models.BooleanField(
        _(u"External ID is set automatically"), default=False)

    class Meta:
        verbose_name = _(u"Container")
        verbose_name_plural = _(u"Containers")
        ordering = ('cached_label',)
        unique_together = ('index', 'location')

    def __unicode__(self):
        lbl = u"{} ({})".format(self.reference, self.container_type)
        return lbl

    def _generate_cached_label(self):
        items = [self.reference, self.precise_location]
        cached_label = u" | ".join(items)
        return cached_label

    def _generate_cached_location(self):
        items = [self.location.name, unicode(self.index)]
        cached_label = u" - ".join(items)
        return cached_label

    @classmethod
    def get_query_owns(cls, user):
        return Q(history_creator=user) | \
            Q(location__person_in_charge__ishtaruser=user.ishtaruser) | \
            Q(responsible__person_in_charge__ishtaruser=user.ishtaruser)

    @property
    def associated_filename(self):
        filename = datetime.date.today().strftime('%Y-%m-%d')
        filename += u'-' + self.reference
        filename += u"-" + self.location.name
        filename += u"-" + unicode(self.index)
        filename += u"-" + self.divisions_lbl
        return slugify(filename)

    @property
    def precise_location(self):
        return self.location.name + u" - " + self.divisions_lbl

    @property
    def divisions_lbl(self):
        locas = [
            u"{} {}".format(loca.division.division, loca.reference)
            for loca in ContainerLocalisation.objects.filter(
                container=self)
            ]
        return u" | ".join(locas)

    def pre_save(self):
        if not self.index:
            q = Container.objects.filter(responsible=self.responsible).order_by(
                '-index')
            if q.count():
                self.index = q.all()[0].index + 1
            else:
                self.index = 1

    def save(self, *args, **kwargs):
        super(Container, self).save(*args, **kwargs)

        updated = False
        if not self.index:
            self.skip_history_when_saving = True
            q = Container.objects.filter(responsible=self.responsible).order_by(
                '-index')
            if q.count():
                self.index = q.all()[0].index + 1
            else:
                self.index = 1
            updated = True

        self.skip_history_when_saving = True
        if not self.external_id or self.auto_external_id:
            external_id = get_external_id('container_external_id', self)
            if external_id != self.external_id:
                updated = True
                self.auto_external_id = True
                self.external_id = external_id

        if updated:
            self._cached_label_checked = False
            self.save()
            return

        # remove old location in warehouse
        q = ContainerLocalisation.objects.filter(container=self).exclude(
            division__warehouse=self.location)
        for loca in q.all():
            loca.delete()

post_save.connect(cached_label_changed, sender=Container)


class ContainerLocalisation(models.Model):
    container = models.ForeignKey(Container, verbose_name=_(u"Container"),
                                  related_name='division')
    division = models.ForeignKey(WarehouseDivisionLink,
                                 verbose_name=_(u"Division"))
    reference = models.CharField(_(u"Reference"), max_length=200, default='')

    class Meta:
        verbose_name = _(u"Container localisation")
        verbose_name_plural = _(u"Container localisations")
        unique_together = ('container', 'division')
        ordering = ('container', 'division__order')

    def __unicode__(self):
        lbl = u" - ".join((unicode(self.container),
                           unicode(self.division), self.reference))
        return lbl

    def save(self, *args, **kwargs):
        super(ContainerLocalisation, self).save(*args, **kwargs)
        cached_label_changed(Container, instance=self.container)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010-2017 Étienne Loks  <etienne.loks_AT_peacefrogsDOTnet>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# See the file COPYING for details.

from django.conf import settings
from django.contrib.sites.models import Site

from ishtar_common.version import __version__
from ishtar_common.models import get_current_profile

from menus import Menu


def get_base_context(request):
    dct = {'URL_PATH': settings.URL_PATH}
    try:
        dct["APP_NAME"] = Site.objects.get_current().name
    except Site.DoesNotExist:
        dct["APP_NAME"] = settings.APP_NAME
    dct["COUNTRY"] = settings.COUNTRY
    """
    if 'MENU' not in request.session or \
       request.session['MENU'].user != request.user:
        menu = Menu(request.user)
        menu.init()
        request.session['MENU'] = menu
    """  # temporary disabled
    current_action = None
    if 'CURRENT_ACTION' in request.session:
        dct['CURRENT_ACTION'] = request.session['CURRENT_ACTION']
        current_action = dct['CURRENT_ACTION']
    dct['CURRENT_PATH'] = request.path
    menu = Menu(request.user, current_action=current_action,
                session=request.session)
    menu.init()
    if menu.selected_idx is not None:
        dct['current_theme'] = "theme-%d" % (menu.selected_idx + 1)
    request.session['MENU'] = menu
    dct['MENU'] = request.session['MENU']
    dct['JQUERY_URL'] = settings.JQUERY_URL
    dct['JQUERY_UI_URL'] = settings.JQUERY_UI_URL
    dct['COUNTRY'] = settings.COUNTRY
    dct['VERSION'] = __version__
    if settings.EXTRA_VERSION:
        dct['VERSION'] += "-" + unicode(settings.EXTRA_VERSION)
    profile = get_current_profile()
    if profile:
        dct['raw_css'] = profile.get_raw_css()
    return dct
